"""Microapp compiler inspector"""

import os, shutil

from collections import Mapping

from microapp import App, register_appclass, unregister_appclass

def _update(d, u):
    for k, v in u.items():
        if isinstance(v, Mapping):
            r = _update(d.get(k, {}), v)
            d[k] = r
        else:
            if k in d:
                if isinstance( u[k], int ):
                    d[k] += u[k]
                else:
                    d[k] = u[k]
            else:
                d[k] = u[k]
    return d


class FortranTimingCollector(App):

    _name_ = "timingcollect"
    _version_ = "0.1.0"

    def __init__(self, mgr):

        self.add_argument("datadir", help="raw data directory")
        self.add_argument("-o", "--output", type=str, help="output path.")

        self.register_forward("data", help="timing raw data")

    def perform(self, args):

        datadir = args.datadir["_"]

        # collect data
        etimes = {} # mpirank:omptid:invoke=[(fileid, linenum, numvisits), ... ]
        etimemin = 1.0E100
        etimemax = 0.0
        netimes = 0
        etimeresol = 0.0
        nexcluded_under = 0
        nexcluded_over = 0

        mpipaths = []
        for item in os.listdir(datadir):
            try:
                mpirank, ompthread = item.split('.')
                if mpirank.isdigit() and ompthread.isdigit():
                    mpipaths.append((datadir, mpirank, ompthread))
            except:
                pass

        nprocs = 1
        fwds = {"data": mpipaths}
        group_opts = ["--multiproc", "%d" % nprocs, "--assigned-input",
               "data:@data", "--clone", "%d" % len(mpipaths), "--data-join", "accumulate"]
        #group_opts = ["--assigned-input", "data:@data", "--clone", "%d" % len(mpipaths), "--forwarding", "accumulate"]
        app_args = ["_tcollect", "@data"]

        register_appclass(_TCollect)

        #ret, fwds = self.manager.run_command(cmd, forward=fwds)
        ret, fwds = self.run_subgroup(group_args=group_opts, app_args=app_args, forward=fwds)
        assert ret == 0, "_tcollect returned non-zero code."

        unregister_appclass(_TCollect)

        # how many groups per app
        for grp, pathouts in fwds.items():
            # how many exits per group
            for pathout in pathouts:
                # how many edges per exit
                for etime, emeta in pathout["data"]: # using group command
                #etime, emeta = pathout["data"]

                    _update(etimes, etime)

                    etimemin = min(etimemin, emeta[0])
                    etimemax = max(etimemax, emeta[1])
                    netimes += emeta[2]
                    etimeresol = max(etimeresol, emeta[3])
                    nexcluded_under += emeta[4]
                    nexcluded_over += emeta[5]

        if len(etimes) == 0:
            shutil.rmtree(datadir)

        etimes["_summary_"] = {
                "elapsedtime_min": etimemin,
                "elapsedtime_max": etimemax,
                "number_eitmes": netimes,
                "elapsedtime_res": etimeresol,
                "number_underflow": nexcluded_under,
                "number_overflow": nexcluded_over
            }

        self.add_forward(data=etimes)

        
class _TCollect(App):

    _name_ = "_tcollect"
    _version_ = "0.1.0"

    def __init__(self, mgr):

        self.add_argument("data", help="input data")
        self.add_argument("-o", "--output", type=str, help="output path.")

        self.register_forward("data", help="timing data")

    def perform(self, args):

        # collect data
        etimes = {} # mpirank:omptid:invoke=[(fileid, linenum, numvisits), ... ]
        emeta = [ 1.0E100, 0.0, 0, 0.0, 0, 0 ] #  min, max, number, resolution, under limit, over limit

#        etimemin_limit = Config.model['types']['etime']['minval']
#        etimemax_limit = Config.model['types']['etime']['maxval']
        etimemin_limit = None
        etimemax_limit = None

        etimeroot, mpirank, ompthread = args.data["_"]

        try:
            if mpirank not in etimes: etimes[mpirank] = {}
            if ompthread not in etimes[mpirank]: etimes[mpirank][ompthread] = {}

            with open(os.path.join(etimeroot, '%s.%s'%(mpirank, ompthread)), 'r') as f:
                for line in f:
                    invoke, start, stop, resolution = line.split()
                    estart = float(start)
                    estop = float(stop)
                    ediff = estop - estart
                    if etimemin_limit is not None and ediff < etimemin_limit:
                        emeta[4] += 1
                    elif etimemax_limit is not None and ediff > etimemax_limit:
                        emeta[5] += 1
                    else:
                        etimes[mpirank][ompthread][invoke] = ( start, stop )

                        if ediff < emeta[0]:
                            emeta[0] = ediff
                        if ediff > emeta[1]:
                            emeta[1] = ediff
                        emeta[2] += 1
                        eresol = float(resolution)
                        if eresol > emeta[3]:
                            emeta[3] = eresol

        except Exception as e:
            pass
            # TODO log error message
        finally:
            pass

        self.add_forward(data=(etimes, emeta))


class FortranTimingCombiner(App):

    _name_ = "timingcombine"
    _version_ = "0.1.0"

    def __init__(self, mgr):

        self.add_argument("data", help="timing data")
        self.add_argument("-o", "--output", type=str, help="output path.")

        self.register_forward("data", help="timing model data")

    def perform(self, args):

        import pdb; pdb.set_trace()

#        else:
#            try:
#                etime_sections = [ Config.path['etime'], 'summary']
#
#                self.addmodel(Config.path['etime'], etime_sections)
#
#                # elapsedtime section
#                etime = []
#                #fd.write('; <MPI rank> < OpenMP Thread> <invocation order> =  <file number>:<line number><num of invocations> ...\n')
#
#                for ranknum, threadnums in etimes.items():
#                    for threadnum, invokenums in threadnums.items():
#                        for invokenum, evalues  in invokenums.items():
#                            etime.append( ( '%s %s %s'%(ranknum, threadnum, invokenum), ', '.join(evalues) ) )
#                self.addsection(Config.path['etime'], Config.path['etime'], etime)
#
#                summary = []
#                summary.append( ('minimum_elapsedtime', str(etimemin)) )
#                summary.append( ('maximum_elapsedtime', str(etimemax)) )
#                summary.append( ('number_elapsedtimes', str(netimes)) )
#                summary.append( ('resolution_elapsedtime', str(etimeresol)) )
#                self.addsection(Config.path['etime'], 'summary', summary )
#            except Exception as e:
#                kgutils.logger.error(str(e))
#
#
#    out, err, retcode = kgutils.run_shcmd('make recover', cwd=etime_realpath)
#
#    if Config.state_switch['clean']:
#        kgutils.run_shcmd(Config.state_switch['clean'])

########################
#
#        #nprocs = min( len(mpipaths), multiprocessing.cpu_count()*1)
#        nprocs = 1
#
#        if nprocs == 0:
#            kgutils.logger.warn('No elapsedtime data files are found.')
#        else:
#            workload = [ chunk for chunk in chunks(mpipaths, int(math.ceil(len(mpipaths)/nprocs))) ]
#            inqs = []
#            outqs = []
#            for _ in range(nprocs):
#                inqs.append(multiprocessing.Queue())
#                outqs.append(multiprocessing.Queue())
#
#            procs = []
#            for idx in range(nprocs):
#                proc = multiprocessing.Process(target=readdatafiles, args=(inqs[idx], outqs[idx]))
#                procs.append(proc)
#                proc.start()
#
#            for inq, chunk in zip(inqs,workload):
#                inq.put(chunk)
#
#            for outq in outqs:
#                etime, emeta = outq.get()
#                update(etimes, etime)
#                etimemin = min(etimemin, emeta[0])
#                etimemax = max(etimemax, emeta[1])
#                netimes += emeta[2]
#                etimeresol = max(etimeresol, emeta[3])
#                nexcluded_under += emeta[4]
#                nexcluded_over += emeta[5]
#            for idx in range(nprocs):
#                procs[idx].join()
#
#
#            if len(etimes) == 0:
#                if not _DEBUG:
#                    shutil.rmtree(datadir)
#                kgutils.logger.warn('Elapsedtime data is not collected.')
#            else:
#                try:
#                    etime_sections = [ Config.path['etime'], 'summary']
#
#                    self.addmodel(Config.path['etime'], etime_sections)
#
#                    # elapsedtime section
#                    etime = []
#                    #fd.write('; <MPI rank> < OpenMP Thread> <invocation order> =  <file number>:<line number><num of invocations> ...\n')
#
#                    for ranknum, threadnums in etimes.items():
#                        for threadnum, invokenums in threadnums.items():
#                            for invokenum, evalues  in invokenums.items():
#                                etime.append( ( '%s %s %s'%(ranknum, threadnum, invokenum), ', '.join(evalues) ) )
#                    self.addsection(Config.path['etime'], Config.path['etime'], etime)
#
#                    summary = []
#                    summary.append( ('minimum_elapsedtime', str(etimemin)) )
#                    summary.append( ('maximum_elapsedtime', str(etimemax)) )
#                    summary.append( ('number_elapsedtimes', str(netimes)) )
#                    summary.append( ('resolution_elapsedtime', str(etimeresol)) )
#                    self.addsection(Config.path['etime'], 'summary', summary )
#
#                except Exception as e:
#                    kgutils.logger.error(str(e))
#
#        else:
#            if not _DEBUG:
#                shutil.rmtree(datadir)
#            kgutils.logger.info('failed to generate elapsedtime information')
#
#        out, err, retcode = kgutils.run_shcmd('make recover', cwd=etime_realpath)
#
#        if Config.state_switch['clean']:
#            kgutils.run_shcmd(Config.state_switch['clean'])
#    else: # check if coverage should be invoked
#        kgutils.logger.info('Reusing Elapsedtime file: %s/%s'%(Config.path['outdir'], Config.modelfile))
#
#
#
#
#
#
#
#
#
#
#
#
#
#
#
#
#
#
#
#
#
#    # check if elapsedtime data exists in model file
#    if not os.path.exists('%s/%s'%(Config.path['outdir'], Config.modelfile)):
#        kgutils.logger.warn('No elapsedtime file is found.')
#    else:
#        # read ini file
#        kgutils.logger.info('Reading %s/%s'%(Config.path['outdir'], Config.modelfile))
#
#        cfg = configparser.ConfigParser()
#        cfg.optionxform = str
#        cfg.read('%s/%s'%(Config.path['outdir'], Config.modelfile))
#
#        try:
#
#            etimemin = float(cfg.get('elapsedtime.summary', 'minimum_elapsedtime').strip())
#            etimemax = float(cfg.get('elapsedtime.summary', 'maximum_elapsedtime').strip())
#            netimes = int(cfg.get('elapsedtime.summary', 'number_elapsedtimes').strip())
#            etimediff = etimemax - etimemin
#            etimeres = float(cfg.get('elapsedtime.summary', 'resolution_elapsedtime').strip())
#
#            # <MPI rank> < OpenMP Thread> <invocation order> =  <file number>:<line number>:<num etimes> ... 
#            if etimediff == 0:
#                nbins = 1
#            else:
#                nbins = max(min(Config.model['types']['etime']['nbins'], netimes), 2)
#
#            kgutils.logger.info('nbins = %d'%nbins)
#            kgutils.logger.info('etimemin = %f'%etimemin)
#            kgutils.logger.info('etimemax = %f'%etimemax)
#            kgutils.logger.info('etimediff = %f'%etimediff)
#            kgutils.logger.info('netimes = %d'%netimes)
#
#            if nbins > 1:
#                etimebins = [ {} for _ in range(nbins) ]
#                etimecounts = [ 0 for _ in range(nbins) ]
#            else:
#                etimebins = [ {} ]
#                etimecounts = [ 0 ]
#
#            idx = 0
#            for opt in cfg.options('elapsedtime.elapsedtime'):
#                ranknum, threadnum, invokenum = tuple( num for num in opt.split() )
#                start, stop = cfg.get('elapsedtime.elapsedtime', opt).split(',')
#                estart = float(start)
#                eend = float(stop)
#                etimeval = eend - estart
#                if nbins > 1:
#                    binnum = int(math.floor((etimeval - etimemin) / etimediff * (nbins - 1)))
#                else:
#                    binnum = 0
#                etimecounts[binnum] += 1
#            kgutils.logger.info('etimediff = %f'%etimediff)
#            kgutils.logger.info('netimes = %d'%netimes)
#
#            if nbins > 1:
#                etimebins = [ {} for _ in range(nbins) ]
#                etimecounts = [ 0 for _ in range(nbins) ]
#            else:
#                etimebins = [ {} ]
#                etimecounts = [ 0 ]
#
#            idx = 0
#            for opt in cfg.options('elapsedtime.elapsedtime'):
#                ranknum, threadnum, invokenum = tuple( num for num in opt.split() )
#                start, stop = cfg.get('elapsedtime.elapsedtime', opt).split(',')
#                estart = float(start)
#                eend = float(stop)
#                etimeval = eend - estart
#                if nbins > 1:
#                    binnum = int(math.floor((etimeval - etimemin) / etimediff * (nbins - 1)))
#                else:
#                    binnum = 0
#                etimecounts[binnum] += 1
#        triples = []
#        for binnum, etimebin in enumerate(etimebins):
#            bin_triples = []
#            range_begin = binnum*(etimemax-etimemin)/nbins + etimemin if binnum > 0  else etimemin
#            range_end = (binnum+1)*(etimemax-etimemin)/nbins + etimemin if binnum < (nbins-1)  else None
#
#            bunit = 'sec'
#            if range_begin < 1.E-6:
#                bunit = 'usec'
#                range_begin *= 1.E6
#
#            if range_end is None:
#                print 'From bin # %d [ %f (%s) ~ ] %f %% of %d'%(binnum, \
#                    range_begin, bunit, countdist[binnum] * 100, totalcount)
#            else:
#                eunit = 'sec'
#                if range_end < 1.E-6:
#                    eunit = 'usec'
#                    range_end *= 1.E6
#
#                print 'From bin # %d [ %f (%s) ~ %f (%s) ] %f %% of %d'%(binnum, \
#                    range_begin, bunit, range_end, eunit, countdist[binnum] * 100, totalcount)
#
#            for invokenum in sorted(etimebin.keys()):
#                if len(bin_triples) >= datacollect[binnum]: break
#                # select datacollect[binum] under this data tree, rank/thread/invoke
#                bininvokes = etimebin[invokenum].keys()
#                random.shuffle(bininvokes)
#                for ranknum in bininvokes:
#                    if len(bin_triples) >= datacollect[binnum]: break
#                    binranks = etimebin[invokenum][ranknum].keys()
#                    random.shuffle(binranks)
#                    for threadnum in binranks:
#                        bin_triples.append( (ranknum, threadnum, invokenum) )
#                        print '        invocation triple: %s:%s:%s'%(ranknum, threadnum, invokenum)
#            triples.extend(bin_triples)
#
#        print 'Number of bins: %d'%nbins
#        print 'Minimun elapsed time: %f'%etimemin
#        print 'Maximum elapsed time: %f'%etimemax
#        #print 'Selected invocation triples:'
#        #print ','.join([ ':'.join([ str(n) for n in t ]) for t in triples])
#
#        for ranknum, threadnum, invokenum in triples:
#            Config.invocation['triples'].append( ( (str(ranknum), str(ranknum)), (str(threadnum), str(threadnum)), \
#                (str(invokenum), str(invokenum)) ) )
#



# gen_write_type.py
from __future__ import absolute_import
 
from fortlab.resolver import statements, block_statements, typedecl_statements, Fortran2003
from fortlab.kgplugin import Kgen_Plugin

from .gencore_utils import get_dtype_writename, get_typedecl_writename, state_gencore_contains, \
    get_dtype_readname, get_typedecl_readname, kernel_gencore_contains, gen_write_istrue, \
    gen_read_istrue, is_excluded, is_remove_state

class Gen_Type(Kgen_Plugin):
    def __init__(self):
        self.frame_msg = None

        self.kernel_created_use_items = []
        self.kernel_created_public_items = []

        self.state_created_use_items = []
        self.state_created_public_items = []

    # registration
    def register(self, msg):
        self.frame_msg = msg

        # register initial events
        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.STATE, GENERATION_STAGE.BEGIN_PROCESS, \
            block_statements.Type, self.has_state_info, self.create_dtype_write_subr) 

        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.STATE, GENERATION_STAGE.BEGIN_PROCESS, \
            statements.Use, self.has_dtype_res_path, self.add_writenames_in_use_public) 

        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.BEGIN_PROCESS, \
            block_statements.Type, self.has_state_info, self.create_dtype_read_subr) 

        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.BEGIN_PROCESS, \
            statements.Use, self.has_dtype_res_path, self.add_readnames_in_use_public) 

    def has_state_info(self, node):
        if node.kgen_stmt and 'abstract' not in node.kgen_stmt.specs and \
            hasattr(node.kgen_stmt, 'geninfo') and len(node.kgen_stmt.geninfo)>0:
            return True
        else: return False

    def has_dtype_res_path(self, node):
        if node.kgen_stmt and hasattr(node.kgen_stmt, 'geninfo'):
            if not node.kgen_stmt.isonly: return False
            for gentype, reqlist in node.kgen_stmt.geninfo.items():
                if any(len(req.res_stmts)>0 and isinstance(req.res_stmts[0], block_statements.Type) and \
                    'abstract' not in req.res_stmts[0].specs for uname, req in reqlist):
                    return True
        return False

    def add_readnames_in_use_public(self, node):
        parent = node.kgen_parent

        for gentype, reqlist in node.kgen_stmt.geninfo.items():
            for uname, req in reqlist:
                if len(req.res_stmts)>0 and isinstance(req.res_stmts[0], block_statements.Type):
                    subrname = get_dtype_readname(req.res_stmts[0])

                    # add use
                    checks = lambda n: n.kgen_match_class==statements.Use and n.kgen_stmt and n.kgen_stmt.isonly and \
                        subrname in n.kgen_stmt.items
                    if (id(parent),subrname) not in self.kernel_created_use_items and not part_has_node(parent, USE_PART, checks):
                        attrs = {'name':node.kgen_stmt.name, 'isonly': True, 'items':[subrname]}
                        part_append_genknode(parent, USE_PART, statements.Use, attrs=attrs)
                        self.kernel_created_use_items.append((id(parent),subrname))

                    # add access
                    checks = lambda n: isinstance(n.kgen_stmt, statements.Public) and n.kgen_stmt.items and subrname in n.kgen_stmt.items
                    if (id(parent),subrname) not in self.kernel_created_public_items and isinstance(parent.kgen_stmt, block_statements.Module) and \
                        not part_has_node(parent, DECL_PART, checks):
                        attrs = {'items':[subrname]}
                        part_append_genknode(parent, DECL_PART, statements.Public, attrs=attrs)
                        self.kernel_created_public_items.append((id(parent),subrname))

    def add_writenames_in_use_public(self, node):
        parent = node.kgen_parent

        for gentype, reqlist in node.kgen_stmt.geninfo.items():
            for uname, req in reqlist:
                if len(req.res_stmts)>0 and isinstance(req.res_stmts[0], block_statements.Type):
                    subrname = get_dtype_writename(req.res_stmts[0])

                    # add use
                    checks = lambda n: n.kgen_match_class==statements.Use and n.kgen_stmt and n.kgen_stmt.isonly and \
                        subrname in n.kgen_stmt.items
                    if (id(parent),subrname) not in self.state_created_use_items and not part_has_node(parent, USE_PART, checks):
                        attrs = {'name':node.kgen_stmt.name, 'isonly': True, 'items':[subrname]}
                        part_append_gensnode(parent, USE_PART, statements.Use, attrs=attrs)
                        self.state_created_use_items.append((id(parent),subrname))
                        parent.kgen_stmt.top.used4genstate = True

                    # add access
                    checks = lambda n: isinstance(n.kgen_stmt, statements.Public) and n.kgen_stmt.items and subrname in n.kgen_stmt.items
                    if (id(parent),subrname) not in self.state_created_public_items and isinstance(parent.kgen_stmt, block_statements.Module) and \
                        not part_has_node(parent, DECL_PART, checks):
                        attrs = {'items':[subrname]}
                        part_append_gensnode(parent, DECL_PART, statements.Public, attrs=attrs)
                        self.state_created_public_items.append((id(parent),subrname))
                        parent.kgen_stmt.top.used4genstate = True


    def get_extends(self, node, bag, depth):
        if isinstance(node, Fortran2003.Type_Attr_Spec) and isinstance(node.items[0], str) and node.items[0] == 'EXTENDS':
            bag.append(node.items[1].string)

    def create_read_intrinsic(self, subrobj, entity_name, stmt, var):

        pobj = gen_read_istrue(subrobj, var, 'var%%%s'%entity_name)

        if var.is_allocatable() or var.is_pointer():
            attrs = {'items': ['var%%%s'%entity_name]}
            part_append_genknode(pobj, EXEC_PART, statements.Allocate, attrs=attrs)

        attrs = {'items': ['var%%%s'%entity_name], 'specs': ['UNIT = kgen_unit']}
        part_append_genknode(pobj, EXEC_PART, statements.Read, attrs=attrs)

        if var.is_array() and stmt.is_numeric():
            attrs = {'designator': 'kgen_array_sumcheck', 'items': ['printname // "%%%s"'%entity_name, 'kgen_array_sum', \
                'DBLE(SUM(var%%%s, mask=(var%%%s .eq. var%%%s)))'%(entity_name, entity_name, entity_name), '.TRUE.']}
            part_append_genknode(pobj, EXEC_PART, statements.Call, attrs=attrs)

        if any(match_namepath(pattern, pack_exnamepath(stmt, entity_name), internal=False) for pattern in getinfo('print_var_names')):
            if stmt.is_numeric() and var.is_array():
                attrs = {'items': ['"KGEN DEBUG: DBLE(SUM(" // printname // " %%%s)) = "'%entity_name, 'DBLE(SUM(var%%%s, mask=(var%%%s .eq. var%%%s)))'%(entity_name, entity_name, entity_name)]}
            else:
                attrs = {'items': ['"KGEN DEBUG: " // printname // "%%%s = "'%entity_name, 'var%%%s'%entity_name]}
            part_append_genknode(pobj, EXEC_PART, statements.Write, attrs=attrs)
        else:
            attrs = {'expr': 'PRESENT( printvar ) .AND. printvar'}
            ifobj = part_append_genknode(pobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

            if stmt.is_numeric() and var.is_array():
                attrs = {'items': ['"KGEN DEBUG: DBLE(SUM(" // printname // "%%%s)) = "'%entity_name, 'DBLE(SUM(var%%%s, mask=(var%%%s .eq. var%%%s)))'%(entity_name, entity_name, entity_name)]}
            else:
                attrs = {'items': ['"KGEN DEBUG: " // printname // "%%%s = "'%entity_name, 'var%%%s'%entity_name]}
            part_append_genknode(ifobj, EXEC_PART, statements.Write, attrs=attrs)

    def create_write_intrinsic(self, subrobj, entity_name, stmt, var):

        pobj = gen_write_istrue(subrobj, var, 'var%%%s'%entity_name)

        attrs = {'items': ['var%%%s'%entity_name], 'specs': ['UNIT = kgen_unit']}
        part_append_gensnode(pobj, EXEC_PART, statements.Write, attrs=attrs)

        if any(match_namepath(pattern, pack_exnamepath(stmt, entity_name), internal=False) for pattern in getinfo('print_var_names')):
            if stmt.is_numeric() and var.is_array():
                attrs = {'items': ['"KGEN DEBUG: DBLE(SUM(" // printname // "%%%s)) = "'%entity_name, 'DBLE(SUM(var%%%s, mask=(var%%%s .eq. var%%%s)))'%(entity_name, entity_name, entity_name)]}
            else:
                attrs = {'items': ['"KGEN DEBUG: " // printname // "%%%s = "'%entity_name, 'var%%%s'%entity_name]}
            part_append_gensnode(pobj, EXEC_PART, statements.Write, attrs=attrs)
        else:
            attrs = {'expr': 'PRESENT( printvar ) .AND. printvar'}
            ifobj = part_append_gensnode(pobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

            if stmt.is_numeric() and var.is_array():
                attrs = {'items': ['"KGEN DEBUG: DBLE(SUM(" // printname // "%%%s)) = "'%entity_name, 'DBLE(SUM(var%%%s, mask=(var%%%s .eq. var%%%s)))'%(entity_name, entity_name, entity_name)]}
            else:
                attrs = {'items': ['"KGEN DEBUG: " // printname // "%%%s = "'%entity_name, 'var%%%s'%entity_name]}
            part_append_gensnode(ifobj, EXEC_PART, statements.Write, attrs=attrs)

    def create_read_call(self, subrobj, callname, entity_name, stmt, var):

        attrs = {'expr': 'PRESENT( printvar ) .AND. printvar'}
        ifobj = part_append_genknode(subrobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'designator': callname, 'items': ['var%%%s'%entity_name, 'kgen_unit', 'printname // "%%%s"'%entity_name, '.TRUE.']}
        part_append_genknode(ifobj, EXEC_PART, statements.Call, attrs=attrs)

        part_append_genknode(ifobj, EXEC_PART, statements.Else)

        pstr = '.TRUE.' if any(match_namepath(pattern, pack_exnamepath(stmt, entity_name), internal=False) for pattern in getinfo('print_var_names')) else '.FALSE.'
        attrs = {'designator': callname, 'items': ['var%%%s'%entity_name, 'kgen_unit', 'printname // "%%%s"'%entity_name, pstr]}
        part_append_genknode(ifobj, EXEC_PART, statements.Call, attrs=attrs)

    def create_write_call(self, subrobj, callname, entity_name, stmt, var):

        attrs = {'expr': 'PRESENT( printvar ) .AND. printvar'}
        ifobj = part_append_gensnode(subrobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'designator': callname, 'items': ['var%%%s'%entity_name, 'kgen_unit', 'printname // "%%%s"'%entity_name, '.TRUE.']}
        part_append_gensnode(ifobj, EXEC_PART, statements.Call, attrs=attrs)

        part_append_gensnode(ifobj, EXEC_PART, statements.Else)

        pstr = '.TRUE.' if any(match_namepath(pattern, pack_exnamepath(stmt, entity_name), internal=False) for pattern in getinfo('print_var_names')) else '.FALSE.'
        attrs = {'designator': callname, 'items': ['var%%%s'%entity_name, 'kgen_unit', 'printname // "%%%s"'%entity_name, pstr]}
        part_append_gensnode(ifobj, EXEC_PART, statements.Call, attrs=attrs)

    # process function
    def create_dtype_read_subr(self, node):
        assert node.kgen_stmt, 'None kgen statement'

        subrname = get_dtype_readname(node.kgen_stmt)
        if subrname is None: return

        parent = node.kgen_parent
        checks = lambda n: isinstance(n.kgen_stmt, block_statements.Subroutine) and n.name==subrname
        if  not part_has_node(parent, SUBP_PART, checks):

            checks = lambda n: n.kgen_isvalid and n.kgen_match_class==statements.Contains
            if not parent in kernel_gencore_contains and not part_has_node(parent, CONTAINS_PART, checks):
                part_append_comment(parent, CONTAINS_PART, '')
                part_append_genknode(parent, CONTAINS_PART, statements.Contains)
                part_append_comment(parent, CONTAINS_PART, '')
                kernel_gencore_contains.append(parent)

            part_append_comment(parent, SUBP_PART, 'read state subroutine for %s'%subrname)
            attrs = {'prefix': 'RECURSIVE', 'name': subrname, 'args': ['var', 'kgen_unit', 'printname', 'printvar']}
            subrobj = part_append_genknode(parent, SUBP_PART, block_statements.Subroutine, attrs=attrs)
            part_append_comment(parent, SUBP_PART, '')

            # variable
            attrs = {'type_spec': 'TYPE', 'attrspec': ['INTENT(INOUT)'], 'selector':(None, node.name), 'entity_decls': ['var']}
            part_append_genknode(subrobj, DECL_PART, typedecl_statements.Type, attrs=attrs)

            # kgen_unit
            attrs = {'type_spec': 'INTEGER', 'attrspec': ['INTENT(IN)'], 'entity_decls': ['kgen_unit']}
            part_append_genknode(subrobj, DECL_PART, typedecl_statements.Integer, attrs=attrs)

            # printname
            attrs = {'type_spec': 'CHARACTER', 'attrspec': ['INTENT(IN)'], 'selector':('*', None), 'entity_decls': ['printname']}
            part_append_genknode(subrobj, DECL_PART, typedecl_statements.Character, attrs=attrs)

            # printvar
            attrs = {'type_spec': 'LOGICAL', 'attrspec': ['INTENT(IN)', 'OPTIONAL'], 'selector':(None, None), 'entity_decls': ['printvar']}
            part_append_genknode(subrobj, DECL_PART, typedecl_statements.Character, attrs=attrs)

            # kgen_istrue
            attrs = {'type_spec': 'LOGICAL', 'entity_decls': ['kgen_istrue']}
            part_append_genknode(subrobj, DECL_PART, typedecl_statements.Logical, attrs=attrs)

            attrs = {'type_spec': 'REAL', 'entity_decls': ['kgen_array_sum'], 'selector': (None, '8')}
            part_append_genknode(subrobj, DECL_PART, typedecl_statements.Real, attrs=attrs)

            part_append_comment(subrobj, DECL_PART, '')

            parent_names = []
            getinfo('traverse')(node.kgen_stmt.f2003, self.get_extends, parent_names)

            extends = []
            for parent_typename in parent_names:
                for uname, req in node.kgen_stmt.unknowns.items():
                    if uname.firstpartname()== parent_typename:
                        if len(req.res_stmts)>0:
                            extends.extend(get_part(req.res_stmts[0].genkpair, TYPE_PART))
                            break

            comp_part = get_part(node, TYPE_PART) 
            for item in extends + comp_part:
                if not hasattr(item, 'kgen_stmt'): continue
                if not isinstance(item.kgen_stmt, typedecl_statements.TypeDeclarationStatement): continue

                stmt = item.kgen_stmt
                entity_names = [ get_entity_name(decl) for decl in stmt.entity_decls ]

                for entity_name, entity_decl in zip(entity_names, stmt.entity_decls):

                    if is_remove_state(entity_name, stmt): continue

                    var = stmt.get_variable(entity_name)
                    callname = get_typedecl_readname(stmt, entity_name)

                    if var.is_array():
                        if stmt.is_derived():
                            self.create_read_call(subrobj, callname, entity_name, stmt, var)
                        else: # intrinsic type
                            if var.is_explicit_shape_array():
                                self.create_read_intrinsic(subrobj, entity_name, stmt, var)
                                callname = None
                            else: # implicit array
                                self.create_read_call(subrobj, callname, entity_name, stmt, var)
                    else: # scalar
                        if stmt.is_derived():
                            if var.is_allocatable() or var.is_pointer():
                                self.create_read_call(subrobj, callname, entity_name, stmt, var)
                            else:
                                callname = None
                                for uname, req in stmt.unknowns.items():
                                    if uname.firstpartname()==stmt.name and len(req.res_stmts)>0:
                                        res = req.res_stmts[0]
                                        callname = get_dtype_readname(res)
                                        break
                                if callname is None:
                                    print('WARNING: Can not find Type resolver for %s'%stmt.name)
                                    part_append_comment(subrobj, EXEC_PART, \
                                        'ERROR: "%s" is not resolved. Call statement to read "%s" is not created here.'%\
                                        (stmt.name, stmt.name))
                                else:
                                    self.create_read_call(subrobj, callname, entity_name, stmt, var)
                        else: # intrinsic type
                            self.create_read_intrinsic(subrobj, entity_name, stmt, var)
                            callname = None

                    if node.kgen_stmt.ancestors()[0] != item.kgen_stmt.ancestors()[0] and callname:
                        # add use
                        pnode = node.kgen_parent
                        checks = lambda n: n.kgen_match_class==statements.Use and n.kgen_stmt and n.kgen_stmt.isonly and \
                            callname in n.kgen_stmt.items
                        if (id(pnode),callname) not in self.state_created_use_items and not part_has_node(pnode, USE_PART, checks):
                            attrs = {'name':item.kgen_stmt.ancestors()[0].name, 'isonly': True, 'items':[callname]}
                            part_append_genknode(pnode, USE_PART, statements.Use, attrs=attrs)
                            self.state_created_use_items.append((id(pnode),callname))

                part_append_comment(subrobj, EXEC_PART, '')

            # create public stmt
            if parent.kgen_match_class in [ block_statements.Program, block_statements.Module ]:
                attrs = {'items': [subrname]}
                part_append_genknode(parent, DECL_PART, statements.Public, attrs=attrs)

    # process function
    def create_dtype_write_subr(self, node):
        assert node.kgen_stmt, 'None kgen statement'

        subrname = get_dtype_writename(node.kgen_stmt)
        if subrname is None: return

        parent = node.kgen_parent
        checks = lambda n: isinstance(n.kgen_stmt, block_statements.Subroutine) and n.name==subrname
        if  not part_has_node(parent, SUBP_PART, checks):

            checks = lambda n: n.kgen_match_class==statements.Contains
            if not parent in state_gencore_contains and not part_has_node(parent, CONTAINS_PART, checks):
                part_append_comment(parent, CONTAINS_PART, '')
                part_append_gensnode(parent, CONTAINS_PART, statements.Contains)
                part_append_comment(parent, CONTAINS_PART, '')
                state_gencore_contains.append(parent)

            part_append_comment(parent, SUBP_PART, 'read state subroutine for %s'%subrname)
            attrs = {'prefix': 'RECURSIVE', 'name': subrname, 'args': ['var', 'kgen_unit', 'printname', 'printvar']}
            subrobj = part_append_gensnode(parent, SUBP_PART, block_statements.Subroutine, attrs=attrs)
            part_append_comment(parent, SUBP_PART, '')

            # variable
            attrs = {'type_spec': 'TYPE', 'attrspec': ['INTENT(IN)'], 'selector':(None, node.name), 'entity_decls': ['var']}
            part_append_gensnode(subrobj, DECL_PART, typedecl_statements.Type, attrs=attrs)

            # kgen_unit
            attrs = {'type_spec': 'INTEGER', 'attrspec': ['INTENT(IN)'], 'entity_decls': ['kgen_unit']}
            part_append_gensnode(subrobj, DECL_PART, typedecl_statements.Integer, attrs=attrs)

            # printname
            attrs = {'type_spec': 'CHARACTER', 'attrspec': ['INTENT(IN)'], 'selector':('*', None), 'entity_decls': ['printname']}
            part_append_gensnode(subrobj, DECL_PART, typedecl_statements.Character, attrs=attrs)

            # printvar
            attrs = {'type_spec': 'LOGICAL', 'attrspec': ['INTENT(IN)', 'OPTIONAL'], 'selector':(None, None), 'entity_decls': ['printvar']}
            part_append_gensnode(subrobj, DECL_PART, typedecl_statements.Character, attrs=attrs)

            # kgen_istrue
            attrs = {'type_spec': 'LOGICAL', 'entity_decls': ['kgen_istrue']}
            part_append_gensnode(subrobj, DECL_PART, typedecl_statements.Logical, attrs=attrs)

            attrs = {'type_spec': 'REAL', 'entity_decls': ['kgen_array_sum'], 'selector': (None, '8')}
            part_append_gensnode(subrobj, DECL_PART, typedecl_statements.Real, attrs=attrs)

            part_append_comment(subrobj, DECL_PART, '')

            parent_names = []
            getinfo('traverse')(node.kgen_stmt.f2003, self.get_extends, parent_names)

            extends = []
            for parent_typename in parent_names:
                for uname, req in node.kgen_stmt.unknowns.items():
                    if uname.firstpartname()== parent_typename:
                        if len(req.res_stmts)>0:
                            extends.extend(get_part(req.res_stmts[0].genspair, TYPE_PART))
                            break

            comp_part = get_part(node, TYPE_PART) 
            for item in extends + comp_part:
                if not hasattr(item, 'kgen_stmt'): continue
                if not isinstance(item.kgen_stmt, typedecl_statements.TypeDeclarationStatement): continue

                stmt = item.kgen_stmt
                entity_names = [ get_entity_name(decl) for decl in stmt.entity_decls ]

                for entity_name, entity_decl in zip(entity_names, stmt.entity_decls):

                    if is_remove_state(entity_name, stmt): continue

                    node.kgen_stmt.top.used4genstate = True

                    var = stmt.get_variable(entity_name)
                    callname = get_typedecl_writename(stmt, entity_name)

                    if var.is_array():
                        if stmt.is_derived():
                            self.create_write_call(subrobj, callname, entity_name, stmt, var)
                        else: # intrinsic type
                            if var.is_explicit_shape_array():
                                self.create_write_intrinsic(subrobj, entity_name, stmt, var)
                                callname = None
                            else: # implicit array
                                self.create_write_call(subrobj, callname, entity_name, stmt, var)
                    else: # scalar
                        if stmt.is_derived():
                            if var.is_allocatable() or var.is_pointer():
                                self.create_write_call(subrobj, callname, entity_name, stmt, var)
                            else:
                                callname = None
                                for uname, req in stmt.unknowns.items():
                                    if uname.firstpartname()==stmt.name and len(req.res_stmts)>0:
                                        res = req.res_stmts[0]
                                        callname = get_dtype_writename(res)
                                        break
                                if callname is None:
                                    print('WARNING: Can not find Type resolver for %s'%stmt.name)
                                    part_append_comment(subrobj, EXEC_PART, \
                                        'ERROR: "%s" is not resolved. Call statement to write "%s" is not created here.'%\
                                        (stmt.name, stmt.name))
                                else:
                                    self.create_write_call(subrobj, callname, entity_name, stmt, var)
                        else: # intrinsic type
                            self.create_write_intrinsic(subrobj, entity_name, stmt, var)
                            callname = None

                    if node.kgen_stmt.ancestors()[0] != item.kgen_stmt.ancestors()[0] and callname:
                        # add use
                        pnode = node.kgen_parent
                        checks = lambda n: n.kgen_match_class==statements.Use and n.kgen_stmt and n.kgen_stmt.isonly and \
                            callname in n.kgen_stmt.items
                        if (id(pnode),callname) not in self.state_created_use_items and not part_has_node(pnode, USE_PART, checks):
                            attrs = {'name':item.kgen_stmt.ancestors()[0].name, 'isonly': True, 'items':[callname]}
                            part_append_gensnode(pnode, USE_PART, statements.Use, attrs=attrs)
                            self.state_created_use_items.append((id(pnode),callname))
                            pnode.kgen_stmt.top.used4genstate = True

                part_append_comment(subrobj, EXEC_PART, '')

            # create public stmt
            if parent.kgen_match_class in [ block_statements.Program, block_statements.Module ]:
                attrs = {'items': [subrname]}
                part_append_gensnode(parent, DECL_PART, statements.Public, attrs=attrs)




import os
import random
import json
from fortlab.kgplugin import Kgen_Plugin
from fortlab.resolver import block_statements, statements, typedecl_statements
import collections

RECL = 10
META = 'metadata.json'

# TODO: gen_coverage subroutine should contains MPI communicator if MPI is enabled

class Gen_Coverage_File(Kgen_Plugin):
    def __init__(self):
        self.frame_msg = None
        self.logger = getinfo('logger')
        self.files = None
        self.lines = None
        self.inv_files = None
        self.inv_lines = None

    # registration
    def register(self, msg):

        self.frame_msg = msg

        # read map
        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.NODE_CREATED, \
            getinfo('topblock_stmt'), None, self.read_maps)

        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.BEGIN_PROCESS, \
            block_statements.IfThen, None, self.addstmt_ifthen)
        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.BEGIN_PROCESS, \
            statements.ElseIf, None, self.addstmt_elseif)
        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.BEGIN_PROCESS, \
            statements.Else, None, self.addstmt_else)
        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.BEGIN_PROCESS, \
            statements.Case, None, self.addstmt_case)

        # when finish process
        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.FINISH_PROCESS, \
            getinfo('topblock_stmt'), None, self.add_blockdata)
        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.NODE_CREATED, \
            block_statements.Program, self.is_driver_name, self.show_coverage)

    def is_driver_name(self, node):
        if node.name==getinfo('kernel_driver_name'): return True
        else: return False

    def show_coverage(self, node):

        try:
            part_append_comment(node, DECL_PART, "#IFDEF KGEN_COVERAGE", style="cpp")
            maxlines = max([len(l) for l in self.lines.values()])
            attrs = {'type_spec': 'INTEGER', 'attrspec': [ 'DIMENSION(0:%d, 0:%d)'%(len(self.files)-1, maxlines-1) ], 'entity_decls': ['kgen_visits']}
            part_append_genknode(node, DECL_PART, typedecl_statements.Integer, attrs=attrs)

            attrs = {'items': [ ( 'coverage', ('kgen_visits', ) ) ]}
            part_append_genknode(node, DECL_PART, statements.Common, attrs=attrs)

            part_append_comment(node, DECL_PART, "#ENDIF", style="cpp")

            part_append_comment(node, DECL_PART, '')

            part_append_comment(node, EXEC_PART, "#IFDEF KGEN_COVERAGE", style="cpp")


            attrs = {'items': ['"****************************************************************************"'], 'specs': [ '*', '"(A)"' ]}
            part_append_genknode(node, EXEC_PART, statements.Write, attrs=attrs)

            attrs = {'items': ['"kernel coverage using input files on ""kgen_statefile.lst"""'], 'specs': [ '*', '"(4X,A)"' ]}
            part_append_genknode(node, EXEC_PART, statements.Write, attrs=attrs)

            attrs = {'items': ['"****************************************************************************"'], 'specs': [ '*', '"(A)"' ]}
            part_append_genknode(node, EXEC_PART, statements.Write, attrs=attrs)

            for fid, fpath in self.inv_files.items():

                attrs = {'items': ['""'], 'specs': [ '*', '"(A)"' ]}
                part_append_genknode(node, EXEC_PART, statements.Write, attrs=attrs)

                if len(fpath) > 50:
                    fpath = ".." + fpath[-50:] 
                attrs = {'specs': ['*', '*'], 'items': [ '"In ""%s"" (file id=%s),"'%(fpath,fid) ]}
                part_append_genknode(node, EXEC_PART, statements.Write, attrs=attrs)

                for lid in sorted(self.inv_lines[fid]):
                    lnum = self.inv_lines[fid][lid]
                    attrs = {'specs': ['*', '"(A,I,A)"'], 'items': ['"  Kernel visited "', \
                        'kgen_visits(%s,%s)'%(fid, lid), '" times at line_id=%s (near original line=%s)"'%(lid, lnum)]}
                        #'kgen_visits(%s,%s)'%(fid, lid), '" times near the line # of %s (line id=%s)"'%(lnum, lid)]}
                    part_append_genknode(node, EXEC_PART, statements.Write, attrs=attrs)

            attrs = {'items': ['"****************************************************************************"'], 'specs': [ '*', '"(A)"' ]}
            part_append_genknode(node, EXEC_PART, statements.Write, attrs=attrs)

            part_append_comment(node, EXEC_PART, "#ENDIF", style="cpp")
        except:
            attrs = {'items': ['"ERROR: Coverage information is disabled due to an internal error."'], 'specs': [ '*', '"(A)"' ]}
            part_append_genknode(node, EXEC_PART, statements.Write, attrs=attrs)

        #import pdb; pdb.set_trace()

    ##################################
    # reading coverage maps
    ##################################

    def read_maps(self, node):

        coverage_files = '%s/__data__/%s/files'%(getinfo('model_path'), getinfo('coverage_typeid'))
        if os.path.exists(coverage_files):
            with open(coverage_files, 'r') as fm:
                self.inv_files = json.load(fm)
                self.files = {}
                for fid, fpath in self.inv_files.items():
                    self.files[fpath] = fid
        else:
            self.logger.warn('Coverage file data does not exist: %s'%coverage_files)

        coverage_lines = '%s/__data__/%s/lines'%(getinfo('model_path'), getinfo('coverage_typeid'))
        if os.path.exists(coverage_lines):
            with open(coverage_lines, 'r') as fm:
                self.inv_lines = json.load(fm)
                self.lines = {}
                for fid, lmap in self.inv_lines.items():
                    self.lines[fid] = linemap = {}
                    for lid, lnum in lmap.items():
                        self.lines[fid][int(lnum)] = int(lid)
        else:
            self.logger.warn('Coverage file data does not exist: %s'%coverage_lines)


    ##################################
    # adding common stmt
    ##################################

    def add_commonstmt(self, node):
        pstmt = node.kgen_stmt.ancestors()[-1]
        pnode = pstmt.genkpair
        if not hasattr(pnode, '__kgen__coverage__common_stmt__'):
            part_append_comment(node, DECL_PART, "#IFDEF KGEN_COVERAGE", style="cpp")
            maxlines = max([len(l) for l in self.lines.values()])
            attrs = {'type_spec': 'INTEGER', 'attrspec': [ 'DIMENSION(0:%d, 0:%d)'%(len(self.files)-1, maxlines-1) ], 'entity_decls': ['kgen_visits']}
            part_append_genknode(pnode, DECL_PART, typedecl_statements.Integer, attrs=attrs)
            attrs = {'items': [ ( 'coverage', ('kgen_visits', ) ) ]}
            part_append_genknode(pnode, DECL_PART, statements.Common, attrs=attrs)
            part_append_comment(node, DECL_PART, "#ENDIF", style="cpp")
            pnode.__kgen__coverage__common_stmt__ = True

    ##################################
    # adding coverage statements
    ##################################

    def ispure(self, node):
        if hasattr(node, 'kgen_stmt') and node.kgen_stmt and \
            isinstance(node.kgen_stmt, block_statements.SubProgramStatement):
            if node.kgen_stmt.is_pure() or node.kgen_stmt.is_elemental():
                return True
        if hasattr(node, 'kgen_parent'):
            return self.ispure(node.kgen_parent)
        return False

    def lineid(self, node):
        if not hasattr(node, 'kgen_stmt') or not hasattr(node.kgen_stmt, 'reader')  or \
            not hasattr(node.kgen_stmt, 'item'):
            return False, False

        path = node.kgen_stmt.reader.id
        lineno = node.kgen_stmt.item.span[0]
        fid = self.files.get(path, False)
        linemap = self.lines.get(fid, {})
        return fid, linemap.get(lineno, False)

    def add_stmt_block(self, node):
        fid, lid = self.lineid(node)
        if fid is not False and lid is not False and not self.ispure(node) and \
            hasattr(node.kgen_stmt, 'unknowns'):
            self.add_commonstmt(node)
            part_insert_comment(node, EXEC_PART, 0, "#IFDEF KGEN_COVERAGE", style="cpp")
            attrs = {'variable': 'kgen_visits(%s, %d)'%(fid, lid), 'sign': '=', 'expr': 'kgen_visits(%s, %d) + 1'%(fid, lid)}
            part_insert_genknode(node, EXEC_PART, statements.Assignment, index=1, attrs=attrs)
            part_insert_comment(node, EXEC_PART, 2, "#ENDIF", style="cpp")

    def add_stmt(self, node):
        fid, lid = self.lineid(node)
        if fid is not False and lid is not False and not self.ispure(node) and \
            hasattr(node.kgen_stmt, 'unknowns'):
            self.add_commonstmt(node)
            idx, name, part = get_part_index(node)
            part_insert_comment(node.kgen_parent, EXEC_PART, idx+1, "#IFDEF KGEN_COVERAGE", style="cpp")
            attrs = {'variable': 'kgen_visits(%s, %d)'%(fid, lid), 'sign': '=', 'expr': 'kgen_visits(%s, %d) + 1'%(fid, lid)}
            part_insert_genknode(node.kgen_parent, EXEC_PART, statements.Assignment, index=(idx+2), attrs=attrs)
            part_insert_comment(node.kgen_parent, EXEC_PART, idx+3, "#ENDIF", style="cpp")

    def addstmt_ifthen(self, node):
        self.add_stmt_block(node)

    def addstmt_elseif(self, node):
        self.add_stmt(node)

    def addstmt_else(self, node):
        self.add_stmt(node)

    def addstmt_case(self, node):
        self.add_stmt(node)

    ##################################
    # adding common block data
    ##################################

    def add_blockdata(self, node):

        maxlines = max([len(l) for l in self.lines.values()])

        part_append_comment(node.kgen_parent, UNIT_PART, '')
        part_append_comment(node.kgen_parent, UNIT_PART, "#IFDEF KGEN_COVERAGE", style="cpp")

        attrs = {'name': 'kgen_coverage_data'}
        cblock = part_append_genknode(node.kgen_parent, UNIT_PART, block_statements.BlockData, attrs=attrs)

        attrs = {'type_spec': 'INTEGER', 'attrspec': [ 'DIMENSION(0:%d, 0:%d)'%(len(self.files)-1, maxlines-1) ], 'entity_decls': ['kgen_visits = 0']}
        part_append_genknode(cblock, DECL_PART, typedecl_statements.Integer, attrs=attrs)

        attrs = {'items': [ ( 'coverage', ('kgen_visits',) ) ]}
        part_append_genknode(cblock, DECL_PART, statements.Common, attrs=attrs)

        part_append_comment(node.kgen_parent, UNIT_PART, "#ENDIF", style="cpp")




# verify_write_typedecl_in_module.py
from __future__ import absolute_import
 
from fortlab.resolver import statements, block_statements, typedecl_statements
from collections import OrderedDict
from fortlab.kgplugin import Kgen_Plugin

from .verify_utils import get_module_verifyname, kernel_verify_contains, VERIFY_PBLOCK_USE_PART, VERIFY_PBLOCK_EXTERNS, \
    get_typedecl_verifyname, get_dtype_verifyname, is_remove_state, is_zero_array, check_class_derived
from .verify_subr import create_verify_subr

class Verify_Typedecl_In_Module(Kgen_Plugin):
    def __init__(self):
        self.frame_msg = None

        self.verify_externs_subrs = OrderedDict()
        self.verify_callsite_use_stmts = []
        self.verify_callsite_call_stmts = []
        self.verify_extern = []
        self.verify_module_subrnames = []

    # registration
    def register(self, msg):
        self.frame_msg = msg

        # register initial events
        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.NODE_CREATED, \
            block_statements.Module, self.has_out_externs_in_module, self.create_verify_module_parts) 

    def is_out_extern_in_verify_module(self, node):
        if node.kgen_stmt and hasattr(node.kgen_stmt, 'geninfo') and len(node.kgen_stmt.geninfo)>0 and \
            KGGenType.has_state_out(node.kgen_stmt.geninfo) and isinstance(node.kgen_parent.kgen_stmt, block_statements.Module):
            return True
        return False

    def has_out_externs_in_module(self, node):
#        checks = lambda n: hasattr(n.kgen_stmt, 'geninfo') and len(n.kgen_stmt.geninfo)>0 \
#            and isinstance(n.kgen_stmt, typedecl_statements.TypeDeclarationStatement) \
#            and KGGenType.has_state_out(n.kgen_stmt.geninfo)
#        if part_has_node(node, DECL_PART, checks):
#            return True
#        return False
        for stmt in node.kgen_stmt.content:
            if isinstance(stmt, typedecl_statements.TypeDeclarationStatement) and \
                hasattr(stmt, 'geninfo') and len(stmt.geninfo)>0 and KGGenType.has_state_out(stmt.geninfo) :
                return True
        return False

    def create_verify_module_parts(self, node):

        attrs = {'name': 'kgen_utils_mod', 'isonly': True, 'items':['check_t', 'kgen_init_check', 'kgen_init_verify',
                'kgen_tolerance', 'kgen_verboselevel', 'kgen_minvalue', 'CHECK_IDENTICAL', 'CHECK_IN_TOL', 'CHECK_OUT_TOL']}
        part_append_genknode(node, USE_PART, statements.Use, attrs=attrs)

        subrname = get_module_verifyname(node.kgen_stmt)
        checks = lambda n: isinstance(n.kgen_stmt, block_statements.Subroutine) and n.name==subrname
        if not part_has_node(node, SUBP_PART, checks):

            checks = lambda n: n.kgen_isvalid and n.kgen_match_class==statements.Contains
            if not node in kernel_verify_contains and not part_has_node(node, CONTAINS_PART, checks):
                part_append_comment(node, CONTAINS_PART, '')
                part_append_genknode(node, CONTAINS_PART, statements.Contains)
                part_append_comment(node, CONTAINS_PART, '')
                kernel_verify_contains.append(node)

            part_append_comment(node, SUBP_PART, 'verify state subroutine for %s'%subrname)
            attrs = {'name': subrname, 'args': ['check_status']}
            subrobj = part_append_genknode(node, SUBP_PART, block_statements.Subroutine, attrs=attrs)
            self.verify_externs_subrs[node] = subrobj

            # check_status
            attrs = {'type_spec': 'TYPE', 'attrspec': ['INTENT(INOUT)'], 'selector':(None, 'check_t'), 'entity_decls': ['check_status']}
            part_append_genknode(subrobj, DECL_PART, typedecl_statements.Type, attrs=attrs)
            part_append_comment(subrobj, DECL_PART, '')

            part_append_comment(node, SUBP_PART, '')

            # add public stmt
            attrs = {'items':[subrname]}
            part_append_genknode(node, DECL_PART, statements.Public, attrs=attrs)

            # register event per typedecl 
            self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.BEGIN_PROCESS, \
                typedecl_statements.TypeDeclarationStatement, self.is_out_extern_in_verify_module, self.create_subr_verify_typedecl_in_module) 

            # register event per module
            self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.BEGIN_PROCESS, \
                block_statements.Module, self.has_out_externs_in_module, self.create_verify_stmts_in_callsite) 

    def create_verify_stmts_in_callsite(self, node):
        if not self.verify_externs_subrs[node] in self.verify_callsite_use_stmts and node.name!=getinfo('topblock_stmt').name:
            attrs = {'name':node.name, 'isonly': True, 'items':[self.verify_externs_subrs[node].name]}
            namedpart_append_genknode(node.kgen_kernel_id, VERIFY_PBLOCK_USE_PART, statements.Use, attrs=attrs)
            self.verify_callsite_use_stmts.append(self.verify_externs_subrs[node])

        if not self.verify_externs_subrs[node] in self.verify_callsite_call_stmts:
            attrs = {'designator': self.verify_externs_subrs[node].name, 'items': ['check_status']}
            namedpart_append_genknode(node.kgen_kernel_id, VERIFY_PBLOCK_EXTERNS, statements.Call, attrs=attrs)
            self.verify_callsite_call_stmts.append(self.verify_externs_subrs[node])

    def create_subr_verify_typedecl_in_module(self, node):
        stmt = node.kgen_stmt
        is_class_derived = check_class_derived(stmt)
        entity_names = set([ uname.firstpartname() for uname, req in KGGenType.get_state_out(stmt.geninfo)])
        for entity_name, entity_decl in zip(entity_names, stmt.entity_decls):
            if entity_name in self.verify_extern: continue
            var = stmt.get_variable(entity_name)

            if is_remove_state(entity_name, stmt): continue
            if var.is_array() and is_zero_array(var, stmt): continue

            self.verify_extern.append(entity_name)
            subrname = get_typedecl_verifyname(stmt, entity_name)

            if subrname not in self.verify_module_subrnames:
                self.verify_module_subrnames.append(subrname)

                if stmt.is_derived():
                    if var.is_pointer() or var.is_array():
                        create_verify_subr(subrname, entity_name, node.kgen_parent, var, stmt)
                    else:
                        subrname = None
                        for uname, req in stmt.unknowns.items():
                            if ( is_class_derived and uname.firstpartname()==stmt.selector[1]) or uname.firstpartname()==stmt.name:
                            #if uname.firstpartname()==stmt.name:
                                if len(req.res_stmts)>0:
                                    subrname = get_dtype_verifyname(req.res_stmts[0])
                                    break
                else:
                    create_verify_subr(subrname, entity_name, node.kgen_parent, var, stmt)

            if subrname:
                attrs = {'designator': subrname, 'items': ['"%s"'%entity_name, 'check_status', entity_name, 'kgenref_%s'%entity_name]}
                part_append_genknode(self.verify_externs_subrs[node.kgen_parent], EXEC_PART, statements.Call, attrs=attrs)





# gen_write_typedecl_in_parentblock.py

import os 
from fortlab.resolver import statements, block_statements, typedecl_statements
from fortlab.kgplugin import Kgen_Plugin

def has_state(node):
    if not hasattr(node, 'kgen_stmt'): return False

    if hasattr(node.kgen_stmt, 'geninfo') or hasattr(node.kgen_stmt, 'unknowns'):
        return True
    else:
        return False

class Debug_Add_Sum_Subprogram(Kgen_Plugin):
    def __init__(self):
        self.frame_msg = None

    # registration
    def register(self, msg):
        self.frame_msg = msg

        # register event per function 
        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.BEGIN_PROCESS, \
            block_statements.Subroutine, has_state, self.add_sum_funcargs) 

        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.STATE, GENERATION_STAGE.BEGIN_PROCESS, \
            block_statements.Subroutine, has_state, self.add_sum_funcargs) 

    def add_sum_funcargs(self, node):
        if not hasattr(node, 'kgen_stmt') or not hasattr(node.kgen_stmt, 'unknowns'): return
        if node.kgen_stmt.is_pure() or node.kgen_stmt.is_elemental(): return

        for uname, req in node.kgen_stmt.unknowns.items():
            if req.res_stmts and req.res_stmts[0]:
                varname = uname.firstpartname()
                stmt = req.res_stmts[0]
                if not isinstance(stmt, typedecl_statements.TypeDeclarationStatement): continue

                var = stmt.get_variable(varname)
                path = os.path.basename(stmt.item.reader.id)
                if stmt.is_numeric() and (var.is_intent_in() or var.is_intent_inout()):
                    node.kgen_stmt.top.used4genstate = True

                    attrs = {'expr': 'kgen_region_started .AND. kgen_collect_log'}
                    ifobj = part_insert_gensnode(node, EXEC_PART, block_statements.IfThen, attrs=attrs, index=0)

                    if var.is_array():
                        attrs = {'items': ['"SUM(%s) at function %s in %s: "'%(varname, node.name, path),'SUM(%s)'%varname]}
                        part_append_gensnode(ifobj, EXEC_PART, statements.Write, attrs=attrs)
                    else:
                        attrs = {'items': ['"%s at function %s in %s: "'%(varname, node.name, path), varname]}
                        part_append_gensnode(ifobj, EXEC_PART, statements.Write, attrs=attrs)


class Debug_Add_Sum_Assignment(Kgen_Plugin):
    def __init__(self):
        self.frame_msg = None

    # registration
    def register(self, msg):
        self.frame_msg = msg

        # register events
        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.NODE_CREATED, \
            block_statements.BeginSource, self.has_topblock, self.create_common_block) 

        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.STATE, GENERATION_STAGE.NODE_CREATED, \
            block_statements.BeginSource, self.has_topblock, self.create_common_block) 

        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.NODE_CREATED, \
            block_statements.Function, None, self.add_common_stmt) 

        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.STATE, GENERATION_STAGE.NODE_CREATED, \
            block_statements.Function, has_state, self.add_common_stmt) 

        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.NODE_CREATED, \
            block_statements.Subroutine, None, self.add_common_stmt) 

        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.STATE, GENERATION_STAGE.NODE_CREATED, \
            block_statements.Subroutine, has_state, self.add_common_stmt) 

        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.NODE_CREATED, \
            statements.Assignment, None, self.add_write_stmt) 

        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.STATE, GENERATION_STAGE.NODE_CREATED, \
            statements.Assignment, has_state, self.add_write_stmt) 

    def has_topblock(self, node):
        checks = lambda n: n.kgen_stmt and n.kgen_stmt==getinfo('topblock_stmt')
        if part_has_node(node, UNIT_PART, checks):
            return True
        else:
            return False

    def create_common_block(self, node):
        # add block data
        attrs = {'name': 'kgen_block_data'}
        bdata = part_append_genknode(node, UNIT_PART, block_statements.BlockData, attrs=attrs)

        attrs = {'type_spec': 'LOGICAL', 'entity_decls': ['kgen_region_started = .FALSE.', 'kgen_collect_log = .FALSE.']}
        part_append_genknode(bdata, DECL_PART, typedecl_statements.Logical, attrs=attrs) 

        attrs = {'items': [('kgen_common_region', ['kgen_region_started', 'kgen_collect_log'])]}
        part_append_genknode(bdata, DECL_PART, statements.Common, attrs=attrs) 

        # register event for topblock file 
        #self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.BEGIN_PROCESS, \
        #    block_statements.BeginSource, self.has_topblock, self.create_common_block) 

        #self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.STATE, GENERATION_STAGE.BEGIN_PROCESS, \
        #    block_statements.BeginSource, self.has_topblock, self.create_common_block) 

    def add_common_stmt(self, node):

        if node.kgen_stmt is None: return

        node.kgen_stmt.top.used4genstate = True

        if not (node.kgen_stmt.is_pure() or node.kgen_stmt.is_elemental()):
            attrs = {'type_spec': 'INTEGER', 'attrspec': ['SAVE'], 'entity_decls': ['kgen_invoc_count = 0']}
            part_append_gensnode(node, DECL_PART, typedecl_statements.Integer, attrs=attrs) 

        attrs = {'type_spec': 'LOGICAL', 'entity_decls': ['kgen_region_started', 'kgen_collect_log']}
        part_append_gensnode(node, DECL_PART, typedecl_statements.Logical, attrs=attrs) 

        attrs = {'items': [('kgen_common_region', ['kgen_region_started', 'kgen_collect_log'])]}
        part_append_gensnode(node, DECL_PART, statements.Common, attrs=attrs)

    def add_write_stmt(self, node):
        if node.kgen_stmt is None: return

        # find typedecl for lhs
        typedecl_stmt = None
        varname = None
        for uname, req in node.kgen_stmt.unknowns.items():
            if node.kgen_stmt.variable.startswith(uname.firstpartname()):
                if isinstance(req.res_stmts[0], typedecl_statements.TypeDeclarationStatement):
                    typedecl_stmt = req.res_stmts[0]
                    varname = uname.firstpartname()
                    break

        if typedecl_stmt and typedecl_stmt.is_numeric():
            var = typedecl_stmt.get_variable(varname)
            path = os.path.basename(node.kgen_stmt.item.reader.id)
            index, partid, part = get_part_index(node)
            upperblock_stmt = node.kgen_stmt.ancestors()[-1]
            if isinstance(upperblock_stmt, block_statements.SubProgramStatement) and (upperblock_stmt.is_pure() or \
                upperblock_stmt.is_elemental()):
                return

            # reduce output log size
            if not node.kgen_stmt.parent is upperblock_stmt: return

            node.kgen_stmt.top.used4genstate = True

            attrs = {'expr': 'kgen_region_started .AND. kgen_collect_log'}
            ifobj = part_insert_gensnode(node.kgen_parent, partid, block_statements.IfThen, attrs=attrs, index=(index+1))

            if var.is_array():
                attrs = {'items': ['"SUM(%s) at function %s in %s: "'%(varname, upperblock_stmt.name, path),'SUM(%s)'%varname]}
                part_append_gensnode(ifobj, EXEC_PART, statements.Write, attrs=attrs)
            else:
                attrs = {'items': ['"%s at function %s in %s: "'%(varname, upperblock_stmt.name, path), varname]}
                part_append_gensnode(ifobj, EXEC_PART, statements.Write, attrs=attrs)





"""
Fortran block statements.

-----
Permission to use, modify, and distribute this software is given under the
terms of the NumPy License. See http://scipy.org.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Author: Pearu Peterson <pearu@cens.ioc.ee>
Created: May 2006
-----
"""

           #'Subroutine','Function','Select','WhereConstruct','ForallConstruct', # KGEN deletion
__all__ = ['BeginSource','Module','PythonModule','Program','BlockData','Interface',
           'Subroutine','Function','SelectCase','SelectType', 'WhereConstruct','ForallConstruct', # KGEN addition
           'IfThen','If','Do','Associate','TypeDecl','Enum',
           'EndSource','EndModule','EndPythonModule','EndProgram','EndBlockData','EndInterface',
           'EndSubroutine','EndFunction','EndSelect','EndWhere','EndForall',
           'EndIfThen','EndDo','EndAssociate','EndType','EndEnum',
           ]

import re
import sys

from fortlab.resolver.base_classes import BeginStatement, EndStatement, Statement,\
     AttributeHolder, ProgramBlock, Variable
from fortlab.resolver.readfortran import Line
from fortlab.resolver.utils import split_comma, filter_stmts, parse_bind, parse_result, AnalyzeError, is_name

#import fortlab.resolver.Fortran2003 as Fortran2003 # KGEN addition
from fortlab.resolver import Fortran2003 # KGEN addition
#import logging # KGEN addition
#logger = logging.getLogger('kgen') # KGEN addition
from fortlab.kgutils import logger


class HasImplicitStmt(object):

    a = AttributeHolder(implicit_rules = {})

    def get_type_by_name(self, name):
        implicit_rules = self.a.implicit_rules
        if implicit_rules is None:
            raise AnalyzeError('Implicit rules mapping is null while getting %r type' % (name))
        l = name[0].lower()
        if l in implicit_rules:
            return implicit_rules[l]
        # default rules:
        if l in 'ijklmn':
            l = 'default_integer'
        else:
            l = 'default_real'
        t = implicit_rules.get(l, None)
        if t is None:
            if l[8:]=='real':
                implicit_rules[l] = t = Real(self, self.item.copy('real'))
            else:
                implicit_rules[l] = t = Integer(self, self.item.copy('integer'))
        return t

    def topyf(self, tab='  '):
        implicit_rules = self.a.implicit_rules
        if implicit_rules is None:
            return tab + 'IMPLICIT NONE\n'
        items = {}
        for c,t in implicit_rules.items():
            if c.startswith('default'):
                continue
            st = t.tostr()
            if st in items:
                items[st].append(c)
            else:
                items[st] = [c]
        if not items:
            return tab + '! default IMPLICIT rules apply\n'
        s = 'IMPLICIT'
        ls = []
        for st,l in items.items():
            l.sort()
            ls.append(st + ' (%s)' % (', '.join(l)))
        s += ' ' + ', '.join(ls)
        return tab + s + '\n'

class HasUseStmt(object):

    a = AttributeHolder(use = {},
                        use_provides = {})

    def get_entity(self, name):
        for modname, modblock in self.top.a.module.items():
            for stmt in modblock.content:
                if getattr(stmt,'name','') == name:
                    return stmt
        return

    def topyf(self, tab='  '):
        sys.stderr.write('HasUseStmt.topyf not implemented\n')
        return ''

class AccessSpecs(object):

    a = AttributeHolder(private_id_list = [], public_id_list = [])

    def topyf(self, tab='  '):
        private_list = self.a.private_id_list
        public_list = self.a.public_id_list
        l = []
        if '' in private_list: l.append(tab + 'PRIVATE\n')
        if '' in public_list: l.append(tab + 'PUBLIC\n')
        for a in private_list:
            if not a: continue
            l.append(tab + 'PRIVATE :: %s\n' % (a))
        for a in public_list:
            if not a: continue
            l.append(tab + 'PUBLIC :: %s\n' % (a))
        return ''.join(l)

class HasVariables(object):

    a = AttributeHolder(variables = {},
                        variable_names = [] # defines the order of declarations
                        )

    def get_variable_by_name(self, name):
        variables = self.a.variables
        if name in variables:
            var = variables[name]
        else:
            var = variables[name] = Variable(self, name)
            self.a.variable_names.append(name)
        return var

    def topyf(self,tab='', only_variables = None):
        s = ''
        if only_variables is None:
            only_variables = self.a.variables.keys()
        for name in only_variables:
            var = self.a.variables[name]
            s += tab + str(var) + '\n'
        return s

class HasTypeDecls(object):

    a = AttributeHolder(type_decls = {})

    def topyf(self, tab=''):
        s = ''
        for name, stmt in self.a.type_decls.items():
            s += stmt.topyf(tab='  '+tab)
        return s

    def get_type_decl_by_kind(self, kind):
        type_decls = self.a.type_decls
        type_decl = type_decls.get(kind, None)
        if type_decl is None:
            return self.get_entity(kind)
        return type_decl

class HasEnumDecls(object):

    a = AttributeHolder(enum_decls = {})

class HasAttributes(object):

    known_attributes = []
    a = AttributeHolder(attributes = [])

    def topyf(self, tab=''):
        s = ''
        for attr in self.a.attributes:
            s += tab + attr + '\n'
        return s

    def update_attributes(self,*attrs):
        attributes = self.a.attributes
        known_attributes = self.known_attributes
        if len(attrs)==1 and isinstance(attrs[0],(tuple,list)):
            attrs = attrs[0]
        for attr in attrs:
            uattr = attr.upper()
            if uattr not in attributes:
                if isinstance(known_attributes,(list, tuple)):
                    if uattr not in known_attributes:
                        self.warning('unknown attribute %r' % (attr))
                elif not known_attributes(uattr):
                    self.warning('unknown attribute %r' % (attr))
                attributes.append(uattr)
            else:
                self.warning('multiple specification of attribute %r' % (attr))                
        return

class HasModuleProcedures(object):

    a = AttributeHolder(module_procedures = [])

# start of KGEN addition
class HasCommonStmts(object):

    a = AttributeHolder(common_stmts = {})

# end of KGEN addition

# File block

class EndSource(EndStatement):
    """
    Dummy End statement for BeginSource.
    """
    match = staticmethod(lambda s: False)

class BeginSource(BeginStatement):
    """
    Fortran source content.
    """
    f2003_class = Fortran2003.Begin_Source # KGEN addition

    match = staticmethod(lambda s: True)
    end_stmt_cls = EndSource

    a = AttributeHolder(module = {},
                        external_subprogram = {},
                        blockdata = {},
                        )
    def tofortran(self,isfix=None):
        if isfix:
            tab = 'C'
        else:
            tab = self.get_indent_tab(isfix=isfix) + '!'
        return tab + BeginStatement.tofortran(self, isfix=isfix)
        
    def tostr(self):
        return self.blocktype.upper() + ' '+ self.name

    # start of KGEN addition
    def tokgen(self):
        return

    # end of KGEN addition

    def process_item(self):
        self.name = self.reader.name
        self.top = self
        self.fill(end_flag = True)
        return

    def analyze(self):
        for stmt in self.content:
            if isinstance(stmt, Module):
                stmt.analyze()
                self.a.module[stmt.name] = stmt
            elif isinstance(stmt, SubProgramStatement):
                stmt.analyze()
                self.a.external_subprogram[stmt.name] = stmt
            elif isinstance(stmt, BlockData):
                stmt.analyze()
                self.a.blockdata[stmt.name] = stmt
            else:
                stmt.analyze()
        return

    def get_classes(self):
        if self.reader.ispyf:
            return [PythonModule] + program_unit
        return program_unit

    def process_subitem(self, item):
        # MAIN block does not define start/end line conditions,
        # so it should never end until all lines are read.
        # However, sometimes F77 programs lack the PROGRAM statement,
        # and here we fix that:
        if self.reader.isf77:
            line = item.get_line()
            if line=='end':
                message = item.reader.format_message(\
                        'WARNING',
                        'assuming the end of undefined PROGRAM statement',
                        item.span[0],item.span[1])
                logger.warning(message)
                # print >> sys.stderr, message
                p = Program(self)
                p.content.extend(self.content)
                p.content.append(EndProgram(p,item))
                self.content[:] = [p]
                return
        return BeginStatement.process_subitem(self, item)

    def topyf(self, tab=''): # XXXX
        s = ''
        for name, stmt in self.a.module.items():
            s += stmt.topyf(tab=tab)
        for name, stmt in self.a.external_subprogram.items():
            s += stmt.topyf(tab=tab)
        for name, stmt in self.a.blockdata.items():
            s += stmt.topyf(tab=tab)
        return s
# Module

class EndModule(EndStatement):
    f2003_class = Fortran2003.End_Module_Stmt # KGEN addition

    match = re.compile(r'end(\s*module\s*\w*|)\Z', re.I).match

class Module(BeginStatement, HasAttributes,
             #HasImplicitStmt, HasUseStmt, HasVariables, # KGEN deletion
             HasImplicitStmt, HasUseStmt, HasVariables, HasCommonStmts, HasEnumDecls, # KGEN addition
             HasTypeDecls, AccessSpecs):
    """
    MODULE <name>
     ..
    END [MODULE [name]]
    """
    f2003_class = Fortran2003.Module_Stmt # KGEN addition

    match = re.compile(r'module\s*\w+\Z', re.I).match
    end_stmt_cls = EndModule

    a = AttributeHolder(module_subprogram = {},
                        module_provides = {}, # all symbols that are public and so
                                              # can be imported via USE statement
                                              # by other blocks
                        #module_interface = {} # KGEN deletion
                        module_interface = [] # KGEN addition
                        )

    known_attributes = ['PUBLIC', 'PRIVATE']

    def get_classes(self):
        return access_spec + specification_part + module_subprogram_part

    def process_item(self):
        name = self.item.get_line().replace(' ','')[len(self.blocktype):].strip()
        self.name = name
        return BeginStatement.process_item(self)

    def get_provides(self):
        return self.a.module_provides

    def get_interface(self):
        return self.a.module_interface

    def analyze(self):
        content = self.content[:]

        while content:
            stmt = content.pop(0)
            if isinstance(stmt, Contains):
                stmt.analyze() # KGEN addition
                for stmt in filter_stmts(content, SubProgramStatement):
                    stmt.analyze()
                    self.a.module_subprogram[stmt.name] = stmt
                stmt = content.pop(0)
                while isinstance(stmt, Comment):
                    stmt = content.pop(0)
                if not isinstance(stmt, EndModule):
                    stmt.error('Expected END MODULE statement (analyzer).')
                continue
            #from statements import Comment
            try:
                stmt.analyze()

            except Exception as e:
                #import pdb ;pdb.set_trace()
                print (e)

        if content:
            logger.debug('Not analyzed content: %s' % content)
            # self.show_message('Not analyzed content: %s' % content)

        module_provides = self.a.module_provides
        for name, var in self.a.variables.items():
            if var.is_public():
                if name in module_provides:
                    self.warning('module data object name conflict with %s, overriding.' % (name))
                module_provides[name] = var
        return

    def topyf(self, tab=''):
        s = tab + 'MODULE '+self.name + '\n'
        s +=  HasImplicitStmt.topyf(self, tab=tab+'  ')
        s +=  AccessSpecs.topyf(self, tab=tab+'  ')
        s +=  HasAttributes.topyf(self, tab=tab+'  ')
        s +=  HasTypeDecls.topyf(self, tab=tab+'  ')
        s +=  HasVariables.topyf(self, tab=tab+'  ')
        #for name, stmt in self.a.module_interface.items(): # KGEN deletion
        for stmt in self.a.module_interface: # KGEN addition
            s += stmt.topyf(tab=tab+'    ')
        s +=  tab + '  CONTAINS\n'
        for name, stmt in self.a.module_subprogram.items():
            s += stmt.topyf(tab=tab+'    ')
        s += tab + 'END MODULE ' + self.name + '\n'
        return s

    def check_private(self, name):
        if name in self.a.public_id_list: return False
        if name in self.a.private_id_list: return True
        if '' in self.a.public_id_list: return False
        if '' in self.a.private_id_list: return True
        #todo: handle generic-spec-s in id-lists.
        return

# Python Module

class EndPythonModule(EndStatement):
    match = re.compile(r'end(\s*python\s*module\s*\w*|)\Z', re.I).match

class PythonModule(BeginStatement, HasImplicitStmt, HasUseStmt,
                   HasVariables):
    """
    PYTHON MODULE <name>
     ..
    END [PYTHON MODULE [name]]
    """
    modes = ['pyf']
    match = re.compile(r'python\s*module\s*\w+\Z', re.I).match
    end_stmt_cls = EndPythonModule

    def get_classes(self):
        return [Interface, Function, Subroutine, Module]

    def process_item(self):
        self.name = self.item.get_line().replace(' ','')\
                    [len(self.blocktype):].strip()
        return BeginStatement.process_item(self)

# Program

class EndProgram(EndStatement):
    """
    END [PROGRAM [name]]
    """
    f2003_class = Fortran2003.End_Program_Stmt # KGEN addition

    match = re.compile(r'end(\s*program\s*\w*|)\Z', re.I).match

class Program(BeginStatement, ProgramBlock,
              HasAttributes, # XXX: why Program needs .attributes? # uncommended KGEN
              #HasVariables, # KGEN deletion
              HasVariables, HasCommonStmts, # KGEN addition
              HasImplicitStmt, HasUseStmt, AccessSpecs):
    """ PROGRAM [name]
    """
    f2003_class = Fortran2003.Program_Stmt # KGEN addition

    a = AttributeHolder(internal_subprogram = {})

    match = re.compile(r'program\s*\w*\Z', re.I).match
    end_stmt_cls = EndProgram

    def get_classes(self):
        return specification_part + execution_part + internal_subprogram_part

    def process_item(self):
        if self.item is not None:
            name = self.item.get_line().replace(' ','')\
                   [len(self.blocktype):].strip()
            if name:
                self.name = name
        return BeginStatement.process_item(self)

    # start of KGEN
    def analyze(self):
        content = self.content[:]

#        if self.prefix:
#            self.update_attributes(self.prefix.upper().split())
#
#        variables = self.a.variables
#        for a in self.args:
#            assert a not in variables
#            if is_name(a):
#                variables[a] = Variable(self, a)
#            elif a=='*':
#                variables[a] = Variable(self, a) # XXX: fix me appropriately
#            else:
#                raise AnalyzeError('argument must be a name or * but got %r' % (a)) 
#
#        if isinstance(self, Function):
#            var = variables[self.result] = Variable(self, self.result)
#            if self.typedecl is not None:
#                var.set_type(self.typedecl)

        while content:
            stmt = content.pop(0)
            if isinstance(stmt, Contains):
                stmt.analyze()
                for stmt in filter_stmts(content, SubProgramStatement):
                    stmt.analyze()
                    self.a.internal_subprogram[stmt.name] = stmt
                stmt = content.pop(0)
                while isinstance (stmt, Comment):
                    stmt = content.pop(0)
                assert isinstance(stmt, self.end_stmt_cls),str(stmt)
            elif isinstance(stmt, self.end_stmt_cls):
                continue
            else:
                stmt.analyze()

        if content:
            logger.debug('Not analyzed content: %s' % content)
            # self.show_message('Not analyzed content: %s' % content)

#        parent_provides = self.parent.get_provides()
#        if parent_provides is not None:
#            if self.name in parent_provides:
#                self.warning('module subprogram name conflict with %s, overriding.' % (self.name))
#            if self.is_public():
#                parent_provides[self.name] = self
#
#        if self.is_recursive() and self.is_elemental():
#            self.warning('C1241 violation: prefix cannot specify both ELEMENTAL and RECURSIVE')
        return
    # end of KGEN
# BlockData

class EndBlockData(EndStatement):
    """
    END [ BLOCK DATA [ <block-data-name> ] ]
    """
    match = re.compile(r'end(\s*block\s*data\s*\w*|)\Z', re.I).match
    #blocktype = 'blockdata' # KGEN deletion
    blocktype = 'block data' # KGEN addition

class BlockData(BeginStatement, HasImplicitStmt, HasUseStmt,
                #HasVariables, AccessSpecs): # KGEN deletion
                HasVariables, AccessSpecs, HasCommonStmts): # KGEN addition
    """
    BLOCK DATA [ <block-data-name> ]
    """
    f2003_class = Fortran2003.Block_Data_Stmt # KGEN addition

    end_stmt_cls = EndBlockData
    match = re.compile(r'block\s*data\s*\w*\Z', re.I).match

    def process_item(self):
        self.name = self.item.get_line()[5:].lstrip()[4:].lstrip()
        return BeginStatement.process_item(self)

    def get_classes(self):
        return specification_part

    # start of KGEN addition
    def tostr(self):
        if hasattr(self, 'name') and self.name:
            return 'BLOCK DATA %s'%self.name
        return 'BLOCK DATA'
    # end of KGEN addition

# Interface

class EndInterface(EndStatement):
    f2003_class = Fortran2003.End_Interface_Stmt # KGEN addition

    match = re.compile(r'end\s*interface\s*(\w+\s*\(.*\)|\w*)\Z',re.I).match
    blocktype = 'interface'

class Interface(BeginStatement, HasAttributes, HasImplicitStmt, HasUseStmt,
                HasModuleProcedures, AccessSpecs
                ):
    """
    INTERFACE [<generic-spec>] | ABSTRACT INTERFACE
    END INTERFACE [<generic-spec>]

    <generic-spec> = <generic-name>
                   | OPERATOR ( <defined-operator> )
                   | ASSIGNMENT ( = )
                   | <dtio-generic-spec>
    <dtio-generic-spec> = READ ( FORMATTED )
                        | READ ( UNFORMATTED )
                        | WRITE ( FORMATTED )
                        | WRITE ( UNFORMATTED )

    """
    f2003_class = Fortran2003.Interface_Stmt # KGEN addition

    modes = ['free', 'fix', 'pyf']
    match = re.compile(r'(interface\s*(\w+\s*\(.*\)|\w*)|abstract\s*interface)\Z',re.I).match
    end_stmt_cls = EndInterface
    blocktype = 'interface'

    a = AttributeHolder(interface_provides = {})

    # start of KGEN
    def resolve(self, request, config):
        from fortlab.resolver.kgparse import ResState
        from fortlab.resolver.kgsearch import f2003_search_unknowns

        if request is None: return

        logger.debug('%s is being resolved'%request.uname.firstpartname())

        # if resolved, return
        if request.state == ResState.RESOLVED:
            logger.debug('%s is already resolved'%request.uname.firstpartname())
            return

        if request.uname.firstpartname()==self.name or \
            ( hasattr(self, 'absinf_procname') and request.uname.firstpartname() == self.absinf_procname ):
            logger.debug('The request is being resolved by an interface')
            request.res_stmt = self
            request.state = ResState.RESOLVED
            request.res_stmt.add_geninfo(request.uname, request)
            #self.check_spec_stmts(request.uname, request.res_stmt)

            for _stmt, _depth in walk(request.res_stmt, -1):
                if not hasattr(_stmt, 'unknowns'):
                    f2003_search_unknowns(_stmt, _stmt.f2003)
                for unk, req in _stmt.unknowns.items():
                    if req.state != ResState.RESOLVED:
                        _stmt.resolve(req, config) 

        # defer to super
        if request.state != ResState.RESOLVED:
            super(Interface, self).resolve(request, config)

    # end of KGEN

    def get_classes(self):
        l = intrinsic_type_spec + interface_specification
        if self.reader.mode=='pyf':
            return [Subroutine, Function] + l
        return l

    def process_item(self):
        line = self.item.get_line()
        line = self.item.apply_map(line)
        self.isabstract = line.startswith('abstract')
        if self.isabstract:
            self.generic_spec = ''
        else:
            self.generic_spec = line[len(self.blocktype):].strip()
        self.name = self.generic_spec # XXX
        return BeginStatement.process_item(self)

    def tostr(self):
        if self.isabstract:
            return 'ABSTRACT INTERFACE'
        return 'INTERFACE '+ str(self.generic_spec)

    #def get_provides(self):
    #    return self.a.interface_provides

    def analyze(self):
        content = self.content[:]

        absinf_subp_processed = False # KGEN addition

        while content:
            stmt = content.pop(0)
            if isinstance(stmt, self.end_stmt_cls):
                break

            if self.isabstract and not absinf_subp_processed and isinstance(stmt, SubProgramStatement):
                self.absinf_procname = stmt.name

            stmt.analyze()
            #assert isinstance(stmt, SubProgramStatement),`stmt.__class__.__name__`
        if content:
            logger.debug('Not analyzed content: %s' % content)
            # self.show_message('Not analyzed content: %s' % content)

        if self.name in self.parent.a.variables:
            var = self.parent.a.variables.pop(self.name)
            self.update_attributes(var.attributes)

        #if isinstance(self.parent, Module):# KGEN deletion
        if isinstance(self.parent, ( SubProgramStatement, Module )): # KGEN addition
            parent_interface = self.parent.get_interface()
            # start of KGEN deletion
#            if self.name in parent_interface:
#                p = parent_interface[self.name]
#                last = p.content.pop()
#                assert isinstance(last,EndInterface),`last.__class__`
#                p.content += self.content
#                p.update_attributes(self.a.attributes)
#            else:
#                parent_interface[self.name] = self
            parent_interface.append(self)
            # end of KGEN deletion
            return

    def topyf(self, tab=''):
        s = tab + self.tostr() + '\n'
        s +=  HasImplicitStmt.topyf(self, tab=tab+'  ')
        s +=  HasAttributes.topyf(self, tab=tab+'  ')
        s +=  HasUseStmt.topyf(self, tab=tab+'  ')
        s += tab + 'END' + self.tostr() + '\n'
        return s

# Subroutine

class SubProgramStatement(BeginStatement, ProgramBlock,
                          HasImplicitStmt, HasAttributes,
                          #HasUseStmt, # KGEN deletion
                          HasUseStmt, HasCommonStmts, HasEnumDecls, # KGEN addition
                          HasVariables, HasTypeDecls, AccessSpecs
                          ):
    """
    [ <prefix> ] <FUNCTION|SUBROUTINE> <name> [ ( <args> ) ] [ <suffix> ]
    """

    #a = AttributeHolder(internal_subprogram = {}) # KGEN deletion
    a = AttributeHolder(internal_subprogram = {}, subprogram_interface = []) # KGEN addition
    known_attributes = ['RECURSIVE', 'PURE', 'ELEMENTAL']


    def tofortran(self, isfix=None):
        construct_name = self.construct_name
        construct_name = construct_name + ': ' if construct_name else ''
        l=[self.get_indent_tab(isfix=isfix) + construct_name + self.tostr()]
        for c in self.content:
            l.append(c.tofortran(isfix=isfix))
        return '\n'.join(l)

    # start of KGEN
    def tokgen(self):
        construct_name = self.construct_name
        construct_name = construct_name + ': ' if construct_name else ''

        tosubr = False
        if hasattr(self, 'tosurb'):
            tosubr = True
            del self.tosurb

        if tosubr: clsname = 'SUBROUTINE'
        elif isinstance(self, Statement): clsname = self.__class__.__name__.upper()
        else: clsname = self.kgen_match_class.__name__.upper()

        if hasattr(self, 'new_args'):
            args = self.new_args
            del self.new_args
        elif hasattr(self, 'args'): args = self.args
        else: args = []

        s = ''
        if hasattr(self, 'prefix') and self.prefix:
            s += self.prefix + ' '
        if not tosubr:
            if isinstance(self, Statement):
                if hasattr(self, 'typedecl') and self.typedecl is not None:
                    assert isinstance(self, Function),str(self.__class__.__name__)
                    if not hasattr(self, 'result_in_typedecl') or not self.result_in_typedecl:
                        s += self.typedecl.tostr() + ' '
                elif hasattr(self.parent, 'funcresult_in_stmt') and self.name in self.parent.funcresult_in_stmt.keys():
                    s += self.parent.funcresult_in_stmt[self.name] + ' '
        s += clsname
        suf = ''
        if not tosubr:
            if hasattr(self, 'result') and self.result and self.result!=self.name:
                suf += ' RESULT ( %s )' % (self.result)
        if hasattr(self, 'bind') and self.bind:
            suf += ' BIND ( %s )' % (', '.join(self.bind))

        return construct_name + '%s %s(%s)%s' % (s, self.name,', '.join(args),suf) 

    def get_interface(self):
        return self.a.subprogram_interface

    # end of KGEN

    def process_item(self):
        clsname = self.__class__.__name__.lower()
        item = self.item
        line = item.get_line()
        m = self.match(line)
        i = line.lower().find(clsname)
        assert i!=-1,str(clsname, line)
        self.prefix = line[:i].rstrip()
        self.name = line[i:m.end()].lstrip()[len(clsname):].strip()
        line = line[m.end():].lstrip()
        args = []
        if line.startswith('('):
            i = line.find(')')
            assert i!=-1,str(line)
            line2 = item.apply_map(line[:i+1])
            for a in line2[1:-1].split(','):
                a=a.strip()
                if not a: continue
                args.append(a)
            line = line[i+1:].lstrip()
        suffix = item.apply_map(line)
        self.bind, suffix = parse_bind(suffix, item)
        self.result = None
        if isinstance(self, Function):
            self.result, suffix = parse_result(suffix, item)
            if suffix:
                assert self.bind is None,str(self.bind)
                #self.bind, suffix = parse_result(suffix, item) # KGEN deletion
                self.bind, suffix = parse_bind(suffix, item) # KGEN addition
            if self.result is None:
                self.result = self.name
        assert not suffix,str(suffix)
        self.args = args
        self.typedecl = None
        return BeginStatement.process_item(self)

    def tostr(self):
        clsname = self.__class__.__name__.upper()

        s = ''
        if self.prefix: # KGEN deletion
            s += self.prefix + ' '
        if self.typedecl is not None: # KGEN deletion
            # start of KGEN addtion
            if hasattr(self, 'typedecl') and self.typedecl is not None:
                assert isinstance(self, Function),str(self.__class__.__name__)
                if not hasattr(self, 'result_in_typedecl') or not self.result_in_typedecl:
                    s += self.typedecl.tostr() + ' '
            elif hasattr(self.parent, 'funcresult_in_stmt') and self.name in self.parent.funcresult_in_stmt.keys():
                s += self.parent.funcresult_in_stmt[self.name] + ' '
            # end of KGEN addtion
            #s += self.typedecl.tostr() + ' ' # KGEN deletion
        s += clsname
        suf = ''
        #if self.result and self.result!=self.name: # KGEN deletion
        if self.result and self.result!=self.name: # KGEN addition
            suf += ' RESULT ( %s )' % (self.result)
        if self.bind:
            suf += ' BIND ( %s )' % (', '.join(self.bind))

        return '%s %s(%s)%s' % (s, self.name,', '.join(self.args),suf) # KGEN deletion

    def get_classes(self):
        return f2py_stmt + specification_part + execution_part \
               + internal_subprogram_part

    def analyze(self):
        content = self.content[:]

        if self.prefix:
            self.update_attributes(self.prefix.upper().split())

        variables = self.a.variables
        for a in self.args:
            assert a not in variables
            if is_name(a):
                variables[a] = Variable(self, a)
            elif a=='*':
                variables[a] = Variable(self, a) # XXX: fix me appropriately
            else:
                raise AnalyzeError('argument must be a name or * but got %r' % (a)) 

        if isinstance(self, Function):
            var = variables[self.result] = Variable(self, self.result)
            if self.typedecl is not None:
                var.set_type(self.typedecl)

        while content:
            stmt = content.pop(0)
            if isinstance(stmt, Contains):
                stmt.analyze() # KGEN addition
                for stmt in filter_stmts(content, SubProgramStatement):
                    stmt.analyze()
                    self.a.internal_subprogram[stmt.name] = stmt
                stmt = content.pop(0)
                while isinstance (stmt, Comment):
                    stmt = content.pop(0)
                assert isinstance(stmt, self.end_stmt_cls),str(stmt)
            elif isinstance(stmt, self.end_stmt_cls):
                continue
            else:
                stmt.analyze()

        if content:
            logger.debug('Not analyzed content: %s' % content)
            # self.show_message('Not analyzed content: %s' % content)

        parent_provides = self.parent.get_provides()
        if parent_provides is not None:
            if self.name in parent_provides:
                self.warning('module subprogram name conflict with %s, overriding.' % (self.name))
            if self.is_public():
                parent_provides[self.name] = self

        if self.is_recursive() and self.is_elemental():
            self.warning('C1241 violation: prefix cannot specify both ELEMENTAL and RECURSIVE')
        return

    def topyf(self, tab=''):
        s = tab + self.__class__.__name__.upper()
        s += ' ' + self.name + ' (%s)' % (', '.join(self.args))
        if isinstance(self, Function) and self.result != self.name:
            s += ' RESULT (%s)' % (self.result)
        s += '\n'
        s +=  HasImplicitStmt.topyf(self, tab=tab+'  ')
        s +=  AccessSpecs.topyf(self, tab=tab+'  ')
        s +=  HasTypeDecls.topyf(self, tab=tab+'  ')
        s +=  HasVariables.topyf(self, tab=tab+'  ', only_variables = self.args)
        s += tab + 'END ' + self.__class__.__name__.upper() + ' ' + self.name + '\n'
        return s

    def is_public(self): return not self.is_private()

    def is_private(self): return self.parent.check_private(self.name)
    def is_recursive(self): return 'RECURSIVE' in self.a.attributes
    def is_pure(self): return 'PURE' in self.a.attributes
    def is_elemental(self): return 'ELEMENTAL' in self.a.attributes

class EndSubroutine(EndStatement):
    """
    END [SUBROUTINE [name]]
    """
    f2003_class = Fortran2003.End_Subroutine_Stmt # KGEN addition

    match = re.compile(r'end(\s*subroutine\s*\w*|)\Z', re.I).match


class Subroutine(SubProgramStatement):
    """
    [ <prefix> ] SUBROUTINE <name> [ ( [ <dummy-arg-list> ] ) [ <proc-language-binding-spec> ]]
    """
    f2003_class = Fortran2003.Subroutine_Stmt # KGEN addition

    end_stmt_cls = EndSubroutine
    match = re.compile(r'(recursive|pure|elemental|\s)*subroutine\s*\w+', re.I).match
    _repr_attr_names = ['prefix','bind','suffix','args'] + Statement._repr_attr_names

# Function

class EndFunction(EndStatement):
    """
    END [FUNCTION [name]]
    """
    f2003_class = Fortran2003.End_Function_Stmt # KGEN addition

    match = re.compile(r'end(\s*function\s*\w*|)\Z', re.I).match

    def tokgen(self):
        tosubr = False
        if hasattr(self, 'tosubr'):
            tosubr = True
            del self.tosubr

        if tosubr: subp_name = 'SUBROUTINE'
        else: subp_name = 'FUNCTION'
        return 'END %s %s'%(subp_name,self.name or '')

class Function(SubProgramStatement):
    """
    [ <prefix> ] FUNCTION <name> ( [<dummy-arg-list>] ) [<suffix>]
    <prefix> = <prefix-spec> [ <prefix-spec> ]...
    <prefix-spec> = <declaration-type-spec>
                  | RECURSIVE | PURE | ELEMENTAL
    <suffix> = <proc-language-binding-spec> [ RESULT ( <result-name> ) ]
             | RESULT ( <result-name> ) [ <proc-language-binding-spec> ]
    """
    f2003_class = Fortran2003.Function_Stmt # KGEN addition

    end_stmt_cls = EndFunction
    match = re.compile(r'(recursive|pure|elemental|\s)*function\s*\w+', re.I).match
    _repr_attr_names = ['prefix','bind','suffix','args','typedecl'] + Statement._repr_attr_names

    def subroutine_wrapper_code(self):
        name = 'f2pywrap_' + self.name
        args = ['f2pyvalue_'+self.result] + self.args
        var = self.a.variables[self.result]
        typedecl = var.get_typedecl().astypedecl()
        lines = []
        tab = ' '*6
        lines.append('%sSUBROUTINE %s(%s)' % (tab, name, ', '.join(args)))
        if isinstance(self.parent,Module):
            lines.append('%s  USE %s' % (tab, self.parent.name))
        else:
            if isinstance(typedecl, TypeStmt):
                type_decl = typedecl.get_type_decl(typedecl.name)
                if type_decl.parent is self:
                    for line in str(type_decl).split('\n'):
                        lines.append('%s  %s' % (tab, line.lstrip()))
            lines.append('%s  EXTERNAL %s' % (tab, self.name))
            lines.append('%s  %s %s' % (tab, str(typedecl).lstrip(), self.name))
        lines.append('%s  %s %s' % (tab, str(typedecl).lstrip(), args[0]))
        lines.append('!f2py intent(out) %s' % (args[0]))
        for a in self.args:
            v = self.a.variables[a]
            lines.append('%s  %s' % (tab, str(v).lstrip()))
        lines.append('%s  %s = %s(%s)' % (tab, args[0], self.name, ', '.join(self.args)))
        #lines.append('%s  print*,"%s=",%s' % (tab, args[0], args[0])) # debug line
        lines.append('%sEND SUBROUTINE %s' % (tab, name))
        return '\n'.join(lines)

    def subroutine_wrapper(self):
        code = self.subroutine_wrapper_code()
        from api import parse
        block = parse(code) # XXX: set include_dirs
        while len(block.content)==1:
            block = block.content[0]
        return block

# Handle subprogram prefixes

class SubprogramPrefix(Statement):
    """
    <prefix> <declaration-type-spec> <function|subroutine> ...
    """
    match = re.compile(r'(pure|elemental|recursive|\s)+\b',re.I).match
    def process_item(self):
        line = self.item.get_line()
        m = self.match(line)
        prefix = line[:m.end()].rstrip()
        rest = self.item.get_line()[m.end():].lstrip()
        if rest:
            self.parent.put_item(self.item.copy(prefix))
            self.item.clone(rest)
            self.isvalid = False
            return
        if self.parent.__class__ not in [Function, Subroutine]:
            self.isvalid = False
            return
        prefix = prefix + ' ' + self.parent.prefix
        self.parent.prefix = prefix.strip()
        self.ignore = True
        return

# SelectCase

class EndSelect(EndStatement):
    f2003_class = Fortran2003.End_Select_Stmt # KGEN addition

    match = re.compile(r'end\s*select\s*\w*\Z', re.I).match
    blocktype = 'select'

#class Select(BeginStatement): # KGEN deletion
class SelectCase(BeginStatement): # KGEN addition
    """
    [ <case-construct-name> : ] SELECT CASE ( <case-expr> )

    """
    f2003_class = Fortran2003.Select_Case_Stmt # KGEN addition

    match = re.compile(r'select\s*case\s*\(.*\)\Z',re.I).match
    end_stmt_cls = EndSelect
    name = ''
    def tostr(self):
        return 'SELECT CASE ( %s )' % (self.expr)
    def process_item(self):
        self.expr = self.item.get_line()[6:].lstrip()[4:].lstrip()[1:-1].strip()
        self.construct_name = self.item.name
        return BeginStatement.process_item(self)

    def get_classes(self):
        return [Case] + execution_part_construct

# start of KGEN addition
# SelectType

class SelectType(BeginStatement):
    """
    [ <select-construct-name> : ] SELECT TYPE ( [ <associate-name> => ] <selector> )

    """
    f2003_class = Fortran2003.Select_Type_Stmt

    match = re.compile(r'select\s*type\s*\(.*\)\Z',re.I).match
    end_stmt_cls = EndSelect
    name = ''
    def tostr(self):
        return 'SELECT TYPE ( %s )' % (self.expr)
    def process_item(self):
        self.expr = self.item.get_line()[6:].lstrip()[4:].lstrip()[1:-1].strip()
        self.construct_name = self.item.name
        return BeginStatement.process_item(self)

    def get_classes(self):
        return [TypeGuard] + execution_part_construct

# end of KGEN addition

# Where

class EndWhere(EndStatement):
    """
    END WHERE [ <where-construct-name> ]
    """
    f2003_class = Fortran2003.End_Where_Stmt # KGEN addition

    match = re.compile(r'end\s*\where\s*\w*\Z',re.I).match


class Where(BeginStatement):
    """
    [ <where-construct-name> : ] WHERE ( <mask-expr> )
    <mask-expr> = <logical-expr>
    """
    f2003_class = Fortran2003.Where_Construct_Stmt # KGEN addition

    match = re.compile(r'where\s*\([^)]*\)\Z',re.I).match
    end_stmt_cls = EndWhere
    name = ''
    def tostr(self):
        return 'WHERE ( %s )' % (self.expr)
    def process_item(self):
        self.expr = self.item.apply_map(self.item.get_line()[5:].lstrip()[1:-1].strip()) # KGEN addition
        #self.expr = self.item.get_line()[5:].lstrip()[1:-1].strip() # KGEN deletion
        self.construct_name = self.item.name
        return BeginStatement.process_item(self)

    def get_classes(self):
        return [Assignment, WhereStmt,
                WhereConstruct, ElseWhere
                ]

    # start of KGEN addition
    def analyze(self):
        pass
    # end of KGEN addition

WhereConstruct = Where

# Forall

class EndForall(EndStatement):
    """
    END FORALL [ <forall-construct-name> ]
    """
    f2003_class = Fortran2003.End_Forall_Stmt # KGEN addition

    match = re.compile(r'end\s*forall\s*\w*\Z',re.I).match

class Forall(BeginStatement):
    """
    [ <forall-construct-name> : ] FORALL <forall-header>
      [ <forall-body-construct> ]...
    <forall-body-construct> = <forall-assignment-stmt>
                            | <where-stmt>
                            | <where-construct>
                            | <forall-construct>
                            | <forall-stmt>
    <forall-header> = ( <forall-triplet-spec-list> [ , <scalar-mask-expr> ] )
    <forall-triplet-spec> = <index-name> = <subscript> : <subscript> [ : <stride> ]
    <subscript|stride> = <scalar-int-expr>
    <forall-assignment-stmt> = <assignment-stmt> | <pointer-assignment-stmt>
    """
    f2003_class = Fortran2003.Forall_Construct_Stmt # KGEN addition

    end_stmt_cls = EndForall
    match = re.compile(r'forall\s*\(.*\)\Z',re.I).match
    name = ''
    def process_item(self):
        self.specs = self.item.get_line()[6:].lstrip()[1:-1].strip()
        return BeginStatement.process_item(self)
    def tostr(self):
        return 'FORALL (%s)' % (self.specs)
    def get_classes(self):
        return [GeneralAssignment, WhereStmt, WhereConstruct,
                ForallConstruct, ForallStmt]

ForallConstruct = Forall

# IfThen

class EndIfThen(EndStatement):
    """
    END IF [ <if-construct-name> ]
    """
    f2003_class = Fortran2003.End_If_Stmt # KGEN addition

    match = re.compile(r'end\s*if\s*\w*\Z', re.I).match
    blocktype = 'if'

class IfThen(BeginStatement):
    """
    [<if-construct-name> :] IF ( <scalar-logical-expr> ) THEN

    IfThen instance has the following attributes:
      expr
    """
    f2003_class = Fortran2003.If_Then_Stmt # KGEN addition

    match = re.compile(r'if\s*\(.*\)\s*then\Z',re.I).match
    end_stmt_cls = EndIfThen
    name = ''

    def tostr(self):
        return 'IF (%s) THEN' % (self.expr)

    def process_item(self):
        item = self.item
        line = item.get_line()[2:-4].strip()
        assert line[0]=='(' and line[-1]==')',str(line)
        self.expr = item.apply_map(line[1:-1].strip())
        self.construct_name = item.name
        return BeginStatement.process_item(self)

    def get_classes(self):
        return [Else, ElseIf] + execution_part_construct

class If(BeginStatement):
    """
    IF ( <scalar-logical-expr> ) action-stmt
    """
    f2003_class = Fortran2003.If_Stmt # KGEN addition

    match = re.compile(r'if\s*\(',re.I).match

#        item = self.item
#        line = item.get_line()[2:-4].strip()
#        assert line[0]=='(' and line[-1]==')',`line`
#        self.expr = item.apply_map(line[1:-1].strip())
#        self.construct_name = item.name
#        return BeginStatement.process_item(self)

    def process_item(self):
        item = self.item
        mode = self.reader.mode
        classes = self.get_classes()
        classes = [cls for cls in classes if mode in cls.modes]

        line = item.get_line()[2:].lstrip()
        # start of KGEN addition
        anyallmatch = re.match(r'^\(\s*(any|all)\s*\(', line)
        if anyallmatch:
            j = line.find(')')
            if j < 0:
                i = -1
            else:
                k = line[j+1:].find(')')
                if k < 0:
                    i = -1
                else:
                    i = line[j+1:].find(')') + j + 1
        else:
            i = line.find(')')
        # end of KGEN addition
        #i = line.find(')') # KGEN deletion
        expr = line[1:i].strip()
        line = line[i+1:].strip()
        if line.lower()=='then':
            self.isvalid = False
            return
        self.expr = item.apply_map(expr)

        if not line:
            newitem = self.get_item()
        else:
            newitem = item.copy(line, apply_map=True)
        newline = newitem.get_line()
        for cls in classes:
            if cls.match(newline):
                stmt = cls(self, newitem)
                if stmt.isvalid:
                    self.content.append(stmt)
                    return
        if not line:
            self.put_item(newitem)
        self.isvalid = False
        return

    def tostr(self):
        assert len(self.content)==1,str(self.content)

        # start of KGEN addtion
        line = str(self.content[0]).lstrip()

        # remove label
        line = self.remove_label(line)
        line = self.remove_construct_name(line)

        return 'IF (%s) %s' % (self.expr, line)
        # end of KGEN addtion
        #return 'IF (%s) %s' % (self.expr, str(self.content[0]).lstrip()) # KGEN deletion


    def tofortran(self,isfix=None):
        return self.get_indent_tab(isfix=isfix) + self.tostr()

    # start of KGEN addtion
    def tokgen(self):
        if hasattr(self, 'action_stmt'):
            line = self.action_stmt
        elif hasattr(self, 'content') and len(self.content)==1:
            line = line = str(self.content[0]).lstrip()
        else:
            raise Exception('Wrong If statement syntac')

        # remove label
        if hasattr(self, 'remove_label'):
            line = self.remove_label(line)
        if hasattr(self, 'self.remove_construct_name'):
            line = self.self.remove_construct_name(line)

        return 'IF (%s) %s' % (self.expr, line)
    # end of KGEN addtion

    def get_classes(self):
        return action_stmt

# Do

class EndDo(EndStatement):
    """
    END DO [ <do-construct-name> ]
    """
    f2003_class = Fortran2003.End_Do_Stmt # KGEN addition

    match = re.compile(r'end\s*do\s*\w*\Z', re.I).match
    blocktype = 'do'

class Do(BeginStatement):
    """
    [ <do-construct-name> : ] DO label [loopcontrol]
    [ <do-construct-name> : ] DO [loopcontrol]

    """
    f2003_class = Fortran2003.Do_Stmt # KGEN addition

    match = re.compile(r'do\b\s*\d*',re.I).match
    item_re = re.compile(r'do\b\s*(?P<label>\d*)\s*,?\s*(?P<loopcontrol>.*)\Z',re.I).match
    end_stmt_cls = EndDo
    name = ''

    def tostr(self):
        l = ['DO']
    # start of KGEN addition
        if hasattr(self, 'endlabel') and self.endlabel:
            l.append(str(self.endlabel))
        if hasattr(self, 'loopcontrol') and self.loopcontrol:
            l.append(str(self.loopcontrol))
    # end of KGEN addition
    # start of KGEN deletion
#        for part in [self.endlabel, self.loopcontrol]:
#            if part:
#                l.append(str(part))
    # end of KGEN deletion
        return ' '.join(l)

    def process_item(self):
        item = self.item
        line = item.get_line()
        m = self.item_re(line)
        label = m.group('label').strip() or None
        if label:
            label = int(label)
        self.endlabel = label
        self.construct_name = item.name
        # start of KGEN addition
        loopcontrol = m.group('loopcontrol').strip()
        if loopcontrol is None or loopcontrol.startswith('='):
            self.isvalid = False
            return
        self.loopcontrol = loopcontrol
        # end of KGEN addition
        #self.loopcontrol = m.group('loopcontrol').strip() # KGEN deletion
        return BeginStatement.process_item(self)

    def process_subitem(self, item):
        r = False
        if self.endlabel:
            label = item.label
            if label == self.endlabel:
                r = True
                if isinstance(self.parent, Do) and label==self.parent.endlabel:
                    # the same item label may be used for different block ends
                    self.put_item(item)
        return BeginStatement.process_subitem(self, item) or r

    def get_classes(self):
        return execution_part_construct

# Associate

class EndAssociate(EndStatement):
    """
    END ASSOCIATE [ <associate-construct-name> ]
    """
    f2003_class = Fortran2003.End_Associate_Stmt # KGEN addition

    match = re.compile(r'end\s*associate\s*\w*\Z',re.I).match

class Associate(BeginStatement):
    """
    [ <associate-construct-name> : ] ASSOCIATE ( <association-list> )
      <block>

    <association> = <associate-name> => <selector>
    <selector> = <expr> | <variable>
    """
    f2003_class = Fortran2003.Associate_Stmt # KGEN addition

    match = re.compile(r'associate\s*\(.*\)\Z',re.I).match
    end_stmt_cls = EndAssociate

    def process_item(self):
        line = self.item.get_line()[9:].lstrip()
        self.associations = line[1:-1].strip()
        return BeginStatement.process_item(self)
    def tostr(self):
        return 'ASSOCIATE (%s)' % (self.associations)
    def get_classes(self):
        return execution_part_construct

# Type

class EndType(EndStatement):
    """
    END TYPE [<type-name>]
    """
    f2003_class = Fortran2003.End_Type_Stmt # KGEN addition

    match = re.compile(r'end\s*type\s*\w*\Z', re.I).match
    blocktype = 'type'

class Type(BeginStatement, HasVariables, HasAttributes, AccessSpecs):
    """
    TYPE [ [ , <type-attr-spec-list>] :: ] <type-name> [ ( <type-param-name-list> ) ]
    <type-attr-spec> = <access-spec> | EXTENDS ( <parent-type-name> )
                       | ABSTRACT | BIND(C)
    """
    f2003_class = Fortran2003.Derived_Type_Stmt # KGEN addition

    match = re.compile(r'type\b\s*').match
    end_stmt_cls = EndType

    a = AttributeHolder(extends = None,
                        parameters = {},
                        component_names = [], # specifies component order for sequence types
                        components = {}
                        )
    known_attributes = re.compile(r'\A(PUBLIC|PRIVATE|SEQUENCE|ABSTRACT|BIND\s*\(.*\))\Z',re.I).match

    def process_item(self):
        line = self.item.get_line()[4:].lstrip()
        if line.startswith('('):
            self.isvalid = False
            return
        specs = []
        i = line.find('::')
        if i!=-1:
            for s in line[:i].split(','):
                s = s.strip()
                if s: specs.append(s)
            line = line[i+2:].lstrip()
        self.specs = specs
        i = line.find('(')
        if i!=-1:
            self.name = line[:i].rstrip()
            assert line[-1]==')',str(line)
            self.params = split_comma(line[i+1:-1].lstrip())
        else:
            self.name = line
            self.params = []
        if not is_name(self.name):
            self.isvalid = False
            return
        return BeginStatement.process_item(self)

    def tostr(self):
        s = 'TYPE'
        #if self.specs: # KGEN deletion
        if hasattr(self, 'specs') and self.specs: # KGEN addition
            s += ', '.join(['']+self.specs) + ' ::'
        s += ' ' + self.name
        #if self.params: # KGEN deletion
        if hasattr(self, 'params') and self.params: # KGEN addition
            s += ' ('+', '.join(self.params)+')'
        return s

    def get_classes(self):
        return [Integer] + private_or_sequence + component_part +\
               type_bound_procedure_part

    def analyze(self):
        for spec in self.specs:
            i = spec.find('(')
            if i!=-1:
                assert spec.endswith(')'),str(spec)
                s = spec[:i].rstrip().upper()
                n = spec[i+1:-1].strip()
                if s=='EXTENDS':
                    self.a.extends = n
                    continue
                elif s=='BIND':
                    args,rest = parse_bind(spec)
                    assert not rest,str(rest)
                    spec = 'BIND(%s)' % (', '.join(args))
                else:
                    spec = '%s(%s)' % (s,n)
            else:
                spec = spec.upper()
            self.update_attributes(spec)

        component_names = self.a.component_names
        content = self.content[:]
        while content:
            stmt = content.pop(0)
            if isinstance(stmt, self.end_stmt_cls):
                break
            stmt.analyze()

        if content:
            logging.info('Not analyzed content: %s' % content)
            # self.show_message('Not analyzed content: %s' % content)

        parameters = self.a.parameters
        components = self.a.components
        component_names = self.a.component_names
        for name in self.a.variable_names:
            var = self.a.variables[name]
            if name in self.params:
                parameters[name] = var
            else:
                component_names.append(name)
                components[name] = var

        self.parent.a.type_decls[self.name] = self

        parent_provides = self.parent.get_provides()
        if parent_provides is not None:
            if self.is_public():
                if self.name in parent_provides:
                    self.warning('type declaration name conflict with %s, overriding.' % (self.name))
                parent_provides[self.name] = self

        return

    def topyf(self, tab=''):
        s = tab + 'TYPE'
        if self.a.extends is not None:
            s += ', EXTENDS(%s) ::' % (self.a.extends)
        s += ' ' + self.name
        if self.a.parameters:
            s += ' (%s)' % (', '.join(self.a.parameters))
        s += '\n'
        s += AccessSpecs.topyf(self, tab=tab+'  ')
        s += HasAttributes.topyf(self, tab=tab+'  ')
        s += HasVariables.topyf(self, tab=tab+'  ')
        s += tab + 'END TYPE ' + self.name + '\n'
        return s

    # Wrapper methods:

    def get_bit_size(self, _cache={}):
        try:
            return _cache[id(self)]
        except KeyError:
            s = 0
            for name,var in self.a.components.items():
                s += var.get_bit_size()
            _cache[id(self)] = s
        return s

    def is_public(self): return not self.is_private()

    def is_private(self):
        if 'PUBLIC' in self.a.attributes: return False
        if 'PRIVATE' in self.a.attributes: return True
        return self.parent.check_private(self.name)

TypeDecl = Type

# Enum

class EndEnum(EndStatement):
    """
    END ENUM
    """

    f2003_class = Fortran2003.End_Enum_Stmt # KGEN addition

    match = re.compile(r'end\s*enum\Z',re.I).match
    blocktype = 'enum'

class Enum(BeginStatement):
    """
    ENUM , BIND(C)
      <enumerator-def-stmt>
      [ <enumerator-def-stmt> ]...
    """

    f2003_class = Fortran2003.Enum_Def_Stmt # KGEN addition

    blocktype = 'enum'
    end_stmt_cls = EndEnum
    match = re.compile(r'enum\s*,\s*bind\s*\(\s*c\s*\)\Z',re.I).match
    def process_item(self):
        return BeginStatement.process_item(self)

    def get_classes(self):
        return [Enumerator]

    # start of KGEN addition
    def analyze(self):

        content = self.content[:]
        while content:
            stmt = content.pop(0)
            if isinstance(stmt, self.end_stmt_cls):
                break
            stmt.analyze()

        if content:
            logging.info('Not analyzed content: %s' % content)
            # self.show_message('Not analyzed content: %s' % content)

        return

    def tokgen(self):
        return self.tostr()

    def tostr(self):
        return "enum,bind(c)"

    # end of KGEN addition
###################################################

from fortlab.resolver import statements
from fortlab.resolver import typedecl_statements
__all__.extend(statements.__all__)
__all__.extend(typedecl_statements.__all__)

from fortlab.resolver.statements import *
from fortlab.resolver.typedecl_statements import *

f2py_stmt = [Threadsafe, FortranName, Depend, Check, CallStatement,
             CallProtoArgument]

access_spec = [Public, Private]

interface_specification = [Function, Subroutine,
                           ModuleProcedure
                           ]

module_subprogram_part = [ Contains, Function, Subroutine ]

specification_stmt = access_spec + [ Allocatable, Asynchronous, Bind,
    Common, Data, Dimension, Equivalence, External, Intent, Intrinsic,
    Namelist, Optional, Pointer, Protected, Save, Target, Volatile,
    Value, Enumerator ]

intrinsic_type_spec = [ SubprogramPrefix, Integer , Real,
    DoublePrecision, Complex, DoubleComplex, Character, Logical, Byte
    ]

derived_type_spec = [  ]
type_spec = intrinsic_type_spec + derived_type_spec
#declaration_type_spec = intrinsic_type_spec + [ TypeStmt, Class, Procedure ] # TEMP
declaration_type_spec = intrinsic_type_spec + [ TypeStmt, Class ]

type_declaration_stmt = declaration_type_spec

private_or_sequence = [ Private, Sequence ]

component_part = declaration_type_spec + [ ModuleProcedure ]

proc_binding_stmt = [SpecificBinding, GenericBinding, FinalBinding]

type_bound_procedure_part = [Contains, Private] + proc_binding_stmt

kgen_added_action_stmt = [ PointerAssignment, Assignment, Else, ElseIf, Case, TypeGuard, \
    ElseWhere, Read0, Read1 ]# KGEN addition

#R214
action_stmt = [ Allocate, GeneralAssignment, Assign, Backspace, Call, Close,
    Continue, Cycle, Deallocate, Endfile, Exit, Flush, ForallStmt,
    Goto, If, Inquire, Nullify, Open, Print, Read, Return, Rewind,
    Stop, Wait, WhereStmt, Write, ArithmeticIf, ComputedGoto,
    AssignedGoto, Pause ] + kgen_added_action_stmt # KGEN addition
#    AssignedGoto, Pause ] # KGEN deletion
# GeneralAssignment = Assignment + PointerAssignment
# EndFunction, EndProgram, EndSubroutine - part of the corresponding blocks

#executable_construct = [ Associate, Do, ForallConstruct, IfThen, # KGEN deletion
#    Select, WhereConstruct ] + action_stmt # KGEN deletion
executable_construct = [ Associate, Do, ForallStmt, ForallConstruct, IfThen, # KGEN addition
    SelectCase, SelectType, WhereConstruct ] + action_stmt # KGEN addition
#Case, see Select

execution_part_construct = executable_construct + [ Format, Entry,
    Data ]

execution_part = execution_part_construct[:]

#C201, R208
for cls in [EndFunction, EndProgram, EndSubroutine]:
    try: execution_part.remove(cls)
    except ValueError: pass

internal_subprogram = [Function, Subroutine]

internal_subprogram_part = [ Contains, ] + internal_subprogram

declaration_construct = [ TypeDecl, Entry, Enum, Format, Interface,
    Parameter, ModuleProcedure, ] + specification_stmt + \
    type_declaration_stmt + [ StmtFuncStatement ] # KGEN addition
    #type_declaration_stmt # KGEN deletion
# stmt-function-stmt

implicit_part = [ Implicit, Parameter, Format, Entry ]

specification_part = [ Use, Import ] + implicit_part + \
                     declaration_construct


external_subprogram = [Function, Subroutine]

main_program = [Program] + specification_part + execution_part + \
               internal_subprogram_part

program_unit = main_program + external_subprogram + [Module,
                                                     BlockData ]



import os, json

from microapp import App, appdict
from collections import OrderedDict

from fortlab.kgutils import dequote
from fortlab.kggenfile import init_plugins, plugin_config, KERNEL_ID_0

here = os.path.dirname(os.path.realpath(__file__))

class MicroappBuildScanner(App):
    _name_ = "buildscan"
    _version_ = "0.1.0"
 
    def __init__(self, mgr):

        self.add_argument("buildcmd", metavar="build command", help="Software build command")
        self.add_argument("--cleancmd", type=str, help="Software clean command.")
        self.add_argument("--outdir", type=str, help="output directory")
        self.add_argument("--workdir", type=str, help="work directory")
        self.add_argument("--savejson", type=str, help="save data in a josn-format file")
        self.add_argument("--reuse", type=str, help="reuse existing file(s)")
        self.add_argument("--backupdir", type=str, help="saving source files used")
        self.add_argument("--verbose", action="store_true", help="show compilation details")
        self.add_argument("--check", action="store_true", help="check strace return code")

        self.register_forward("data", help="json object")

    def perform(self, args):

        strace_data = None
        if args.reuse and os.path.isfile(args.reuse["_"]):
            with open(args.reuse["_"]) as f:
                strace_data = json.load(f)

        if strace_data:
            self.add_forward(data=strace_data)
            return

        #cmd = ["compileroption", args.buildcmd["_"]]
        opts = [args.buildcmd["_"]]

        if args.cleancmd:
            opts += ["--cleancmd", args.cleancmd["_"]]

        if args.workdir:
            opts += ["--workdir", args.workdir["_"]]

        if args.outdir:
            opts += ["--outdir", args.outdir["_"]]

        if args.savejson:
            opts += ["--savejson", args.savejson["_"]]

        if args.backupdir:
            opts += ["--backupdir", args.backupdir["_"]]

        if args.verbose:
            opts += ["--verbose"]

        if args.check:
            opts += ["--check"]

        ret, fwds = self.run_subapp("compileroption", opts)
        assert ret == 0, "compileroption returned non-zero code."

        self.add_forward(data=fwds["data"])


class MicroappRunScanner(App):
    _name_ = "runscan"
    _version_ = "0.1.0"

    def __init__(self, mgr):

        self.add_argument("analysis", help="analysis object")
        self.add_argument("--cleancmd", type=str, help="Software clean command.")
        self.add_argument("--buildcmd", metavar="build command", help="Software build command")
        self.add_argument("--runcmd", metavar="run command", help="Software run command")
        self.add_argument("--outdir", help="output directory")
        self.add_argument("-w", "--wait", help="wait to complete run")
        self.add_argument("-o", "--output", help="json output file")
        self.add_argument("-s", "--add-scan", action="append", help="add scanning method")
        self.add_argument("--no-cache", action="store_true",
                            help="force to collect timing data")
        self.add_argument("--noreuse-rawdata", dest="noreuse_rawdata", action='store_true',
                          help="Control raw data generation for modeling.")

        self.register_forward("model", help="model data file")
        self.register_forward("analysis", help="tree object")

        # model parameters
        self.config = appdict()

        self.config['modelfile'] = 'model.ini'
        self.config['model'] = OrderedDict()
        self.config['model']['path'] = ""
        self.config['model']['reuse_rawdata'] = True
        self.config['model']['types'] = OrderedDict()
        self.config['model']['types']['code'] = OrderedDict()
        self.config['model']['types']['code']['id'] = '0'
        self.config['model']['types']['code']['name'] = 'code'
        self.config['model']['types']['code']['collector'] = 'codecollect'
        self.config['model']['types']['code']['combiner'] = 'codecombine'
        self.config['model']['types']['code']['percentage'] = 99.9
        self.config['model']['types']['code']['filter'] = None
        self.config['model']['types']['code']['ndata'] = 20
        self.config['model']['types']['code']['enabled'] = False
        self.config['model']['types']['etime'] = OrderedDict()
        self.config['model']['types']['etime']['id'] = '1'
        self.config['model']['types']['etime']['name'] = 'etime'
        self.config['model']['types']['etime']['collector'] = 'timingcollect'
        self.config['model']['types']['etime']['combiner'] = 'timingcombine'
        self.config['model']['types']['etime']['nbins'] = 5
        self.config['model']['types']['etime']['ndata'] = 20
        self.config['model']['types']['etime']['minval'] = None
        self.config['model']['types']['etime']['maxval'] = None
        self.config['model']['types']['etime']['timer'] = None
        self.config['model']['types']['etime']['enabled'] = True
        self.config['model']['types']['papi'] = OrderedDict()
        self.config['model']['types']['papi']['id'] = '2'
        self.config['model']['types']['papi']['name'] = 'papi'
        self.config['model']['types']['papi']['collector'] = 'papicollect'
        self.config['model']['types']['papi']['combiner'] = 'papicombine'
        self.config['model']['types']['papi']['nbins'] = 5
        self.config['model']['types']['papi']['ndata'] = 20
        self.config['model']['types']['papi']['minval'] = None
        self.config['model']['types']['papi']['maxval'] = None
        self.config['model']['types']['papi']['header'] = None
        self.config['model']['types']['papi']['event'] = 'PAPI_TOT_INS'
        self.config['model']['types']['papi']['static'] = None
        self.config['model']['types']['papi']['dynamic'] = None
        self.config['model']['types']['papi']['enabled'] = False

    def perform(self, args):

        self.config.update(args.analysis['_'])

        # generating model raw data
        if args.noreuse_rawdata:
            self.config['model']['reuse_rawdata'] = False if args.noreuse_rawdata else True

        ret, fwds = 0, {}

        outdir = args.outdir["_"] if args.outdir else os.getcwd()
        modeldir = os.path.join(outdir, "model")

        self.config["model"]["path"] = os.path.realpath(modeldir)
 
        if args.cleancmd:
            self.config["cmd_clean"]['cmds'] = dequote(args.cleancmd["_"])

        if args.buildcmd:
            self.config["cmd_build"]['cmds'] = dequote(args.buildcmd["_"])

        if args.runcmd:
            self.config["cmd_run"]['cmds'] = dequote(args.runcmd["_"])

        if not os.path.isfile(args.output["_"]) or args.no_cache:

            mtypes = {u"datatype": u"model", u"datamap": {}, u"collectmap": {},
                      u"combinemap": {}}

            scans = [s["_"] for s in args.add_scan]


            if args.add_scan is None or "timing" in scans:
                self.scan_timing(args, modeldir, mtypes, self.config)

            with open('%s/__data__/modeltypes' % modeldir, 'w') as f:
                json.dump(mtypes, f)

            # TODO: wait if needed

            ret, fwds = self.combine_model(modeldir)
            assert ret == 0, "combine_model returned non-zero code."

            if args.output:
                with open(args.output["_"], "w") as f:
                    json.dump(fwds["model"], f)

            self.add_forward(model=fwds['model'])

        else:
            with open(args.output["_"]) as f:
                self.add_forward(model=json.load(f))

        self.add_forward(analysis=self.config)

    def combine_model(self, modeldir):

        #cmd = ["modelcombine", modeldir]

        #return self.manager.run_command(cmd)
        return self.run_subapp("modelcombine", [modeldir])

    def scan_timing(self, args, modeldir, mtypes, config):

        mtypes["datamap"][self.config['model']['types']['etime']['id']] = \
            self.config['model']['types']['etime']['name']

        mtypes["collectmap"][self.config['model']['types']['etime']['id']] = \
            self.config['model']['types']['etime']['collector']

        mtypes["combinemap"][self.config['model']['types']['etime']['id']] = \
            self.config['model']['types']['etime']['combiner']

        etime_plugindir = os.path.join(here, "timing", "plugins", "gencore")

        init_plugins([KERNEL_ID_0], (('etime.gencore', etime_plugindir),), config)

        plugin_config["current"].update(self.config)

        #cmd = ["timinggen", "@analysis"]
        opts = ["@analysis"]


        if args.cleancmd:
            opts += ["--cleancmd", args.cleancmd["_"]]

        if args.buildcmd:
            opts += ["--buildcmd", args.buildcmd["_"]]

        if args.runcmd:
            opts += ["--runcmd", args.runcmd["_"]]

        if args.outdir:
            opts += ["--outdir", args.outdir["_"]]

        if args.no_cache:
            opts += ["--no-cache"]

        #ret, fwds = self.manager.run_command(cmd, forward={"analysis": config})
        ret, fwds = self.run_subapp("timinggen", opts, forward={"analysis": config})
        assert ret == 0, "timinggen returned non-zero code."

        etimedir = fwds["etimedir"]
        #cmd = "shell 'cd %s; make; make recover' --useenv" % fwds["etimedir"]
        opts = ["'make'" , "--useenv", "--workdir", etimedir]
        ret, fwds = self.run_subapp("shell", opts)
        assert ret == 0, "shell make returned non-zero code: %s" % fwds['stderr']

        opts = ["make recover" , "--useenv", "--workdir", etimedir]
        ret, fwds = self.run_subapp("shell", opts)
        assert ret == 0, "shell 'make recover' returned non-zero code: %s" % fwds['stderr']

        
class MicroappModelCombiner(App):

    _name_ = "modelcombine"
    _version_ = "0.1.0"

    def __init__(self, mgr):

        self.add_argument("modeldir", metavar="raw datadir", help="Raw model data directory")
        self.add_argument("-o", "--output", type=str, help="output path.")

        self.register_forward("model", help="model object")


    def perform(self, args):

        modeldir = args.modeldir["_"]
        datadir = os.path.join(modeldir, "__data__")
        metafile = os.path.join(datadir, "modeltypes")

        model = appdict()

        with open(metafile, 'r') as fm:
            mtypes = json.load(fm)

            for scanid, scanname in mtypes["datamap"].items():
                scandir = os.path.join(datadir, scanid)

                if os.path.isdir(scandir) and len(os.listdir(scandir)):
                    collector = mtypes["collectmap"][scanid]
                    #cmd = collector + " " + scandir
                    #ret, fwds = self.manager.run_command(cmd)
                    ret, fwds = self.run_subapp(collector, [scandir])
                    assert ret == 0, "%s returned non-zero code." % collector
                    assert "data" in fwds, "No data is defined in %s." % collector
                    assert fwds["data"], "No %s data is collected." % collector

                    model[scanname] = fwds['data']
                    #combiner = mtypes["combinemap"][scanid]
                    #cmd = combiner + " @data"
                    #ret, fwds = self.manager.run_command(cmd,
                    #ret, fwds = self.run_subapp(combiner, ["@data"],
                    #                forward={"data": fwds["data"]})

        self.add_forward(model=model)
